"""Configuration of ventilation."""
import dataclasses
import datetime


@dataclasses.dataclass
class StateConfig:
	"""Basic state config."""
	level: int
	display_text: str


@dataclasses.dataclass
class StateConfigWithTimeout(StateConfig):
	"""State config with timeout."""
	timeout: int


@dataclasses.dataclass
class StateConfigLongAbsence(StateConfig):
	"""State config for long absence state."""
	duration: int
	start_time: datetime.time = datetime.time(6)


@dataclasses.dataclass
class VentilationConfig:
	"""Config for ventilation."""
	state_normal: StateConfig = dataclasses.field(default_factory=lambda: StateConfig(1, "Normal"))
	state_hand: StateConfigWithTimeout = dataclasses.field(default_factory=lambda: StateConfigWithTimeout(2, "Hand", 3600))
	state_external: StateConfig = dataclasses.field(default_factory=lambda: StateConfig(2, "External"))
	state_humidity: StateConfig = dataclasses.field(default_factory=lambda: StateConfig(2, "Humidity"))
	state_long_absence: StateConfigLongAbsence = dataclasses.field(default_factory=lambda: StateConfigLongAbsence(2, "LongAbsence", 3600, datetime.time(6)))
	state_after_run: StateConfig = dataclasses.field(default_factory=lambda: StateConfig(2, "After run"))  # only used for Helios ventilation


CONFIG_DEFAULT = VentilationConfig()



"""Configuration of shading objects."""

import dataclasses


@dataclasses.dataclass
class ShadingPosition:
	"""Position of shading object"""
	position: float | bool | None
	slat: float | None = None


@dataclasses.dataclass
class ShadingConfig:
	"""Config for shading objects.

	A value of 0 of manual_timeout will disable the timeout of the manual state
	"""
	pos_auto_open: ShadingPosition
	pos_wind_alarm: ShadingPosition | None = None
	pos_sleeping: ShadingPosition | None = None
	pos_sun_protection: ShadingPosition | None = None
	pos_night_close_summer: ShadingPosition | None = None
	pos_night_close_winter: ShadingPosition | None = None
	pos_door_open: ShadingPosition | None = None
	manual_timeout: int = 0
	door_post_time: int = 5 * 60

	def __post_init__(self) -> None:
		"""Validate configuration.

		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: if configuration is not valid
		"""
		if self.door_post_time in {0, None}:
			self.door_post_time = 1


CONFIG_DEFAULT = ShadingConfig(
	pos_auto_open=ShadingPosition(0, 0),
	pos_wind_alarm=ShadingPosition(0, 0),
	pos_sleeping=ShadingPosition(100, 100),
	pos_sun_protection=ShadingPosition(100, None),
	pos_night_close_summer=None,
	pos_night_close_winter=ShadingPosition(100, 100),
	pos_door_open=ShadingPosition(0, 0),
	manual_timeout=24 * 3600,
	door_post_time=5,
)

CONFIG_DEFAULT_ELEVATION_SLAT_WINTER = [(0, 100), (4, 100), (8, 90), (18, 80), (26, 70), (34, 60), (41, 50), (42, 50), (90, 50), ]
CONFIG_DEFAULT_ELEVATION_SLAT_SUMMER = [(0, 100), (4, 100), (8, 100), (18, 100), (26, 100), (34, 90), (41, 80), (42, 80), (90, 80), ]



"""Light HCL rules."""
import abc
import datetime
import logging

import HABApp

import habapp_rules.actors.config.light_hcl
import habapp_rules.actors.state_observer
import habapp_rules.core.exceptions
import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.core.type_of_day
import habapp_rules.system

LOGGER = logging.getLogger(__name__)


# pylint: disable=no-member
class _HclBase(habapp_rules.core.state_machine_rule.StateMachineRule):
	"""Base class for HCL rules."""

	states = [
		{"name": "Manual"},
		{"name": "Hand", "timeout": 99, "on_timeout": "hand_timeout"},
		{"name": "Auto", "initial": "Init", "children": [
			{"name": "Init"},
			{"name": "HCL"},
			{"name": "Sleep", "initial": "Active", "children": [
				{"name": "Active"},
				{"name": "Post", "timeout": 1, "on_timeout": "post_sleep_timeout"}
			]},
			{"name": "Focus"}
		]}
	]

	trans = [
		{"trigger": "manual_on", "source": ["Auto", "Hand"], "dest": "Manual"},
		{"trigger": "manual_off", "source": "Manual", "dest": "Auto"},

		{"trigger": "hand_on", "source": "Auto", "dest": "Hand"},
		{"trigger": "hand_timeout", "source": "Hand", "dest": "Auto"},

		{"trigger": "sleep_start", "source": ["Auto_HCL", "Auto_Focus"], "dest": "Auto_Sleep"},
		{"trigger": "sleep_end", "source": "Auto_Sleep_Active", "dest": "Auto_Sleep_Post"},
		{"trigger": "post_sleep_timeout", "source": "Auto_Sleep_Post", "dest": "Auto_HCL"},

		{"trigger": "focus_start", "source": ["Auto_HCL", "Auto_Sleep"], "dest": "Auto_Focus"},
		{"trigger": "focus_end", "source": "Auto_Focus", "dest": "Auto_HCL"},
	]

	def __init__(self,
	             name_color: str,
	             name_manual: str,
	             config: habapp_rules.actors.config.light_hcl.LightHclConfig,
	             name_sleep_state: str | None = None,
	             name_focus: str | None = None,
	             name_switch_on: str | None = None,
	             name_state: str | None = None,
	             state_label: str | None = None) -> None:
		"""Init base class.

		:param name_color: Name of openHAB color item (NumberItem)
    	:param name_manual: name of OpenHAB switch item to disable all automatic functions (SwitchItem)
		:param config: config for HCL rule
		:param name_sleep_state: [optional] name of OpenHAB sleeping state item (StringItem)
		:param name_focus: [optional] name of OpenHAB focus state item (SwitchItem)
		:param name_switch_on: name of OpenHAB switch item, which additionally triggers a color update if switched on. This can be used to also trigger an update if the power supply or a single light switched on
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		"""
		self._config = config

		if not name_state:
			name_state = f"H_{name_color}_state"

		habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, name_state, state_label)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, name_color)

		# get items
		self._item_color = HABApp.openhab.items.NumberItem.get_item(name_color)
		self._item_manual = HABApp.openhab.items.SwitchItem.get_item(name_manual)
		self._item_sleep = HABApp.openhab.items.StringItem.get_item(name_sleep_state) if name_sleep_state else None
		self._item_focus = HABApp.openhab.items.SwitchItem.get_item(name_focus) if name_focus else None
		self._item_switch_on = HABApp.openhab.items.SwitchItem.get_item(name_switch_on) if name_switch_on else None

		self._validate_config()
		self._state_observer = habapp_rules.actors.state_observer.StateObserverNumber(name_color, self._cb_hand, value_tolerance=10)

		# init state machine
		self._previous_state = None
		self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(
			model=self,
			states=self.states,
			transitions=self.trans,
			ignore_invalid_triggers=True,
			after_state_change="_update_openhab_state")
		self._set_initial_state()

		self._set_timeouts()

		# set callbacks
		self._item_manual.listen_event(self._cb_manual, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_sleep is not None:
			self._item_sleep.listen_event(self._cb_sleep, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_focus is not None:
			self._item_focus.listen_event(self._cb_focus, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_switch_on is not None:
			self._item_switch_on.listen_event(self._cb_switch_on, HABApp.openhab.events.ItemStateChangedEventFilter())

	def _validate_config(self) -> None:
		"""Validate configuration.

		:raises habapp_rules.common.exceptions.HabAppRulesConfigurationException: if config is not valid.
		"""

		if (self._item_sleep is None) != (self._config.sleep_color is None):
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException("Ether item_sleep and sleep_color must be given or none of them!")

		if (self._item_focus is None) != (self._config.focus_color is None):
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException("Ether item_focus and focus_color must be given or none of them!")

	def _set_timeouts(self) -> None:
		"""Set timeouts."""
		self.state_machine.get_state("Auto_Sleep_Post").timeout = self._config.post_sleep_timeout if self._config.post_sleep_timeout else 1
		self.state_machine.get_state("Hand").timeout = self._config.hand_timeout if self._config.hand_timeout else 0

	def _get_initial_state(self, default_value: str = "") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: if OpenHAB item has a state it will return it, otherwise return the given default value
		"""
		if self._item_manual.is_on():
			return "Manual"
		if self._item_sleep is not None and self._item_sleep.value in (habapp_rules.system.SleepState.PRE_SLEEPING.value, habapp_rules.system.SleepState.SLEEPING.value):
			return "Auto_Sleep"
		if self._item_focus is not None and self._item_focus.is_on():
			return "Auto_Focus"
		return "Auto_HCL"

	def _update_openhab_state(self) -> None:
		"""Update OpenHAB state item and other states.

		This should method should be set to "after_state_change" of the state machine.
		"""
		if self.state != self._previous_state:
			super()._update_openhab_state()
			self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

			self._set_light_color()
			self._previous_state = self.state

	def _set_light_color(self):
		"""Set light color."""
		target_color = None

		if self.state == "Auto_HCL":
			target_color = self._get_hcl_color()
		elif self.state == "Auto_Focus":
			target_color = self._config.focus_color
		elif self.state == "Auto_Sleep_Active":
			target_color = self._config.sleep_color

		if target_color is not None:
			self._state_observer.send_command(target_color)

	def on_enter_Auto_Init(self) -> None:  # pylint: disable=invalid-name
		"""Is called on entering of init state"""
		self._set_initial_state()

	@abc.abstractmethod
	def _get_hcl_color(self) -> float | None:
		"""Get HCL color.

		:return: HCL light color
		"""

	@staticmethod
	def _get_interpolated_value(config_start: tuple[float, float], config_end: tuple[float, float], value: float) -> float:
		"""Get interpolated value

		:param config_start: start config
		:param config_end: end config
		:param value: input value which is the input for the interpolation
		:return: interpolated value
		"""
		fit_m = (config_end[1] - config_start[1]) / (config_end[0] - config_start[0])
		fit_t = config_end[1] - fit_m * config_end[0]

		return fit_m * value + fit_t

	def _cb_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the manual switch has a state change event.

		:param event: trigger event
		"""
		if event.value == "ON":
			self.manual_on()
		else:
			self.manual_off()

	def _cb_hand(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Callback, which is triggered by the state observer if a manual change was detected.

		:param event: original trigger event
		"""
		self._instance_logger.debug("Hand detected")
		self.hand_on()

	def _cb_focus(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the focus switch has a state change event.

		:param event: trigger event
		"""
		if event.value == "ON":
			self.focus_start()
		else:
			self.focus_end()

	def _cb_sleep(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the sleep state has a state change event.

		:param event: trigger event
		"""
		if event.value == habapp_rules.system.SleepState.PRE_SLEEPING.value:
			self.sleep_start()
			if self._item_focus and self._item_focus.is_on():
				self._item_focus.oh_send_command("OFF")
		elif event.value == habapp_rules.system.SleepState.AWAKE.value:
			self.sleep_end()

	def _cb_switch_on(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the switch_on_item has a state change event.

		:param event: trigger event
		"""
		if event.value == "ON" and self.state == "Auto_HCL":
			if (target_color := self._get_hcl_color()) is not None:
				self.run.at(1, self._state_observer.send_command, target_color)


class HclElevation(_HclBase):
	"""Sun elevation based HCL.

		# Items:
		Number    Elevation                     "Elevation [%s]"                {channel="astro:sun:home:position#elevation"}
		Number    HCL_Color_Elevation           "HCL Color Elevation"
		Switch    HCL_Color_Elevation_manual    "HCL Color Elevation manual"

		# Rule init:
		habapp_rules.actors.light_hcl.HclElevation(
			"Elevation",
			"HCL_Color_Elevation",
			"HCL_Color_Elevation_manual",
			config=habapp_rules.actors.config.light_hcl.EXAMPLE_CONFIG_ELEVATION,
		)
	"""

	def __init__(self,
	             name_elevation: str,
	             name_color: str,
	             name_manual: str,
	             config: habapp_rules.actors.config.light_hcl.LightHclConfig,
	             name_sleep_state: str | None = None,
	             name_focus: str | None = None,
	             name_switch_on: str | None = None,
	             name_state: str | None = None,
	             state_label: str | None = None
	             ):
		"""Init sun elevation based HCL rule.

		:param name_elevation: Name of sun elevation openHAB item (NumberItem)
		:param name_color: Name of openHAB color item (NumberItem)
    	:param name_manual: name of OpenHAB switch item to disable all automatic functions (SwitchItem)
		:param config: config for HCL rule
		:param name_sleep_state: [optional] name of OpenHAB sleeping state item (StringItem)
		:param name_focus: [optional] name of OpenHAB focus state item (SwitchItem)
		:param name_switch_on: name of OpenHAB switch item, which additionally triggers a color update if switched on. This can be used to also trigger an update if the power supply or a single light switched on
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		"""
		self._item_elevation = HABApp.openhab.items.NumberItem.get_item(name_elevation)

		_HclBase.__init__(self, name_color, name_manual, config, name_sleep_state, name_focus, name_switch_on, name_state, state_label)

		self._item_elevation.listen_event(self._cb_elevation, HABApp.openhab.events.ItemStateChangedEventFilter())
		self._cb_elevation(None)

	def _get_hcl_color(self) -> float | None:
		"""Get HCL color depending on elevation

		:return: HCL light color
		"""
		elevation = self._item_elevation.value

		if elevation is None:
			return None

		return_value = 0
		if elevation <= self._config.color_config[0][0]:
			return_value = self._config.color_config[0][1]

		elif elevation >= self._config.color_config[-1][0]:
			return_value = self._config.color_config[-1][1]

		else:
			for idx, config_itm in enumerate(self._config.color_config):  # pragma: no cover
				if config_itm[0] <= elevation <= self._config.color_config[idx + 1][0]:
					return_value = self._get_interpolated_value(config_itm, self._config.color_config[idx + 1], elevation)
					break

		return return_value

	def _cb_elevation(self, _: HABApp.openhab.events.ItemStateChangedEvent | None) -> None:
		"""Callback which is called if elevation changed"""
		if self.state == "Auto_HCL" and self._item_elevation.value is not None:
			self._state_observer.send_command(self._get_hcl_color())


class HclTime(_HclBase):
	"""Time based HCL.
		# Items:
		Number    HCL_Color_Time           "HCL Color Time"
		Switch    HCL_Color_Time_manual    "HCL Color Time manual"

		# Rule init:
		habapp_rules.actors.light_hcl.HclTime(
			"HCL_Color_Time",
			"HCL_Color_Time_manual",
			config=habapp_rules.actors.config.light_hcl.EXAMPLE_CONFIG_ELEVATION,
		)
	"""

	def __init__(self,
	             name_color: str,
	             name_manual: str,
	             config: habapp_rules.actors.config.light_hcl.LightHclConfig,
	             name_sleep_state: str | None = None,
	             name_focus: str | None = None,
	             name_switch_on: str | None = None,
	             name_state: str | None = None,
	             state_label: str | None = None
	             ) -> None:
		"""Init time based HCL rule.

		:param name_color: Name of openHAB color item (NumberItem)
    	:param name_manual: name of OpenHAB switch item to disable all automatic functions (SwitchItem)
		:param config: config for HCL rule
		:param name_sleep_state: [optional] name of OpenHAB sleeping state item (StringItem)
		:param name_focus: [optional] name of OpenHAB focus state item (SwitchItem)
		:param name_switch_on: name of OpenHAB switch item, which additionally triggers a color update if switched on. This can be used to also trigger an update if the power supply or a single light switched on
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		"""
		_HclBase.__init__(self, name_color, name_manual, config, name_sleep_state, name_focus, name_switch_on, name_state, state_label)
		self.run.every(None, 300, self._update_color)  # every 5 minutes

	def _one_hour_later(self, current_time: datetime.datetime) -> bool:
		"""Check if today the color values will be shifted one hour later in the evening

		:param current_time: current time
		:return: True if next day is a weekend / holiday day
		"""
		if not self._config.shift_weekend_holiday:
			return False

		if current_time.hour > 12 and (habapp_rules.core.type_of_day.is_holiday(1) or habapp_rules.core.type_of_day.is_weekend(1)):
			return True
		if current_time.hour <= 4 and (habapp_rules.core.type_of_day.is_holiday() or habapp_rules.core.type_of_day.is_weekend()):
			return True
		return False

	def _get_hcl_color(self) -> float | None:
		"""Get HCL color depending on time

		:return: HCL light color
		"""
		current_time = datetime.datetime.now()

		if self._one_hour_later(current_time):
			current_time -= datetime.timedelta(hours=1)

		if current_time.hour < self._config.color_config[0][0]:
			start_config = (self._config.color_config[-1][0] - 24, self._config.color_config[-1][1])
			end_config = self._config.color_config[0]

		elif current_time.hour >= self._config.color_config[-1][0]:
			start_config = self._config.color_config[-1]
			end_config = (self._config.color_config[0][0] + 24, self._config.color_config[0][1])

		else:
			for idx, config_itm in enumerate(self._config.color_config):  # pragma: no cover
				if config_itm[0] <= current_time.hour < self._config.color_config[idx + 1][0]:
					start_config = config_itm
					end_config = self._config.color_config[idx + 1]
					break

		return self._get_interpolated_value(start_config, end_config, current_time.hour + current_time.minute / 60)

	def _update_color(self) -> None:
		"""Callback which is called every 5 minutes"""
		if self.state == "Auto_HCL":
			self._state_observer.send_command(self._get_hcl_color())



"""Module for filter functions / rules."""
import logging

import HABApp

import habapp_rules.core.logger

LOGGER = logging.getLogger(__name__)


class ExponentialFilter(HABApp.Rule):
	"""Rules class to apply a exponential filter to a number value.

	# Items:
	Number    BrightnessValue                       "Brightness Value [%d]"                         {channel="..."}
	Number    BrightnessFiltered                    "Brightness filtered [%d]"
	Number    BrightnessFilteredInstantIncrease     "Brightness filtered instant increase [%d]"

	# Rule init:
	habapp_rules.common.filter.ExponentialFilter("BrightnessValue", "BrightnessFiltered", 60)  # filter constant 1 minute
    habapp_rules.common.filter.ExponentialFilter("BrightnessValue", "BrightnessFilteredInstantIncrease", 600, True)  # filter constant 10 minutes + instant increase
	"""

	def __init__(self, name_raw: str, name_filtered: str, tau: int, instant_increase: bool = False, instant_decrease: bool = False):
		"""Init exponential filter rule.

		:param name_raw: name of raw OpenHAB item (NumberItem)
		:param name_filtered: name of filtered OpenHAB item (NumberItem)
		:param tau: filter time constant in seconds. E.g. step from 0 to 1 | tau = 5 seconds -> after 5 seconds the value will be 0,67
		:param instant_increase: if set to True, increase of input values will not be filtered
		:param instant_decrease: if set to True, decrease of input values will not be filtered
		"""
		HABApp.Rule.__init__(self)

		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, self.rule_name)

		if instant_decrease and instant_decrease:
			self._instance_logger.warning("instant_increase and instant_decrease was set to True. This deactivates the filter completely!")

		self._instant_increase = instant_increase
		self._instant_decrease = instant_decrease

		self._item_raw = HABApp.openhab.items.NumberItem.get_item(name_raw)
		self._item_filtered = HABApp.openhab.items.NumberItem.get_item(name_filtered)

		self._previous_value = self._item_raw.value

		sample_time = tau / 5  # fifth part of the filter time constant
		self._alpha = 0.2  # always 0.2 since we always have the fifth part of the filter time constant
		self.run.every(None, sample_time, self._cb_cyclic_calculate_and_update_output)

		if self._instant_increase or self._instant_decrease:
			self._item_raw.listen_event(self._cb_item_raw, HABApp.openhab.events.ItemStateChangedEventFilter())

		self._instance_logger.debug(f"Successfully created exponential filter for item {self._item_raw.name}")

	def _cb_cyclic_calculate_and_update_output(self) -> None:
		"""Calculate the new filter output and update the filtered item. This must be called cyclic"""
		new_value = self._item_raw.value

		if any(not isinstance(value, (int, float)) for value in (self._previous_value, new_value)):
			self._instance_logger.warning(f"New or previous value is not a number: new_value: {new_value} | previous_value: {self._previous_value}")
			return

		self._send_output(filtered_value := self._alpha * new_value + (1 - self._alpha) * self._previous_value)
		self._previous_value = filtered_value

	def _cb_item_raw(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback which is called if the value of the raw item changed.

		:param event: event which triggered this event
		"""
		if self._previous_value is None or self._instant_increase and event.value > self._previous_value or self._instant_decrease and event.value < self._previous_value:
			self._send_output(event.value)
			self._previous_value = event.value

	def _send_output(self, new_value: float) -> None:
		"""Send output to the OpenHAB item.

		:param new_value: new value which should be sent
		"""
		self._item_filtered.oh_send_command(new_value)



"""Rules to handle sun sensors."""
import dataclasses
import logging

import HABApp

import habapp_rules.common.filter
import habapp_rules.common.hysteresis
import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule

LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass
class SunPositionWindow:
	"""Class for defining min / max values for azimuth and elevation."""
	azimuth_min: float
	azimuth_max: float
	elevation_min: float = 0.0
	elevation_max: float = 90.0

	def __post_init__(self) -> None:
		"""Check if the dataclass was initialized correctly."""
		if self.azimuth_min > self.azimuth_max:
			LOGGER.warning(f"azimuth_min should be smaller than azimuth_max -> min / max will be swapped. Given values: azimuth_min = {self.azimuth_min} | azimuth_max = {self.azimuth_max}")
			min_orig = self.azimuth_min
			max_orig = self.azimuth_max
			self.azimuth_min = max_orig
			self.azimuth_max = min_orig

		if self.elevation_min > self.elevation_max:
			LOGGER.warning(f"elevation_min should be smaller than elevation_max -> min / max will be swapped. Given values: elevation_min = {self.elevation_min} | elevation_max = {self.elevation_max}")
			min_orig = self.elevation_min
			max_orig = self.elevation_max
			self.elevation_min = max_orig
			self.elevation_max = min_orig


class _SensorBase(HABApp.Rule):
	"""Base class for sun sensors."""

	def __init__(self, name_input: str, name_output: str, threshold: str | float, filter_tau: int, filter_instant_increase: bool = True, filter_instant_decrease: bool = False) -> None:
		"""Init of base class for sun sensors.

		:param name_input: name of OpenHAB input item (NumberItem)
		:param name_output: name of OpenHAB output item (SwitchItem)
		:param threshold: threshold for the temperature difference which is supposed that sun is shining. Can be given as float value or name of OpenHAB NumberItem
		:param filter_tau: filter constant for the exponential filter. Default is set to 20 minutes
		:param filter_instant_increase: if set to True, increase of input values will not be filtered
		:param filter_instant_decrease: if set to True, decrease of input values will not be filtered
		"""
		# init HABApp Rule
		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, name_output)

		# init exponential filter
		name_input_exponential_filtered = f"H_{name_input.removeprefix('H_')}_filtered"
		habapp_rules.core.helper.create_additional_item(name_input_exponential_filtered, "Number", name_input_exponential_filtered.replace("_", " "))
		habapp_rules.common.filter.ExponentialFilter(name_input, name_input_exponential_filtered, filter_tau, filter_instant_increase, filter_instant_decrease)

		# init items
		self._item_input_filtered = HABApp.openhab.items.NumberItem.get_item(name_input_exponential_filtered)
		self._item_output = HABApp.openhab.items.SwitchItem.get_item(name_output)
		self._item_threshold = HABApp.openhab.items.NumberItem.get_item(threshold) if isinstance(threshold, str) else None

		# attributes
		self._threshold = self._item_threshold.value if self._item_threshold is not None else threshold

		# callbacks
		self._item_input_filtered.listen_event(self._cb_input_filtered, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_threshold is not None:
			self._item_threshold.listen_event(self._cb_threshold, HABApp.openhab.events.ItemStateChangedEventFilter())

	def _send_output(self, new_value: str) -> None:
		"""Send output if different.

		:param new_value: new value which should be sent
		"""
		if new_value != self._item_output.value:
			self._item_output.oh_send_command(new_value)
			self._instance_logger.debug(f"Set output '{self._item_output.name}' to {new_value}")

	def _cb_input_filtered(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the filtered input value changed.

		:param event: trigger event
		"""
		if self._threshold is None:
			self._instance_logger.debug("The threshold value is None -> Can not check if new value is greater / smaller then the threshold")
			return

		value = "ON" if event.value >= self._threshold else "OFF"
		self._send_output(value)

	def _cb_threshold(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the threshold changed.

		:param event: trigger event
		"""
		self._threshold = event.value


class SensorBrightness(_SensorBase):
	"""Rules class to set sun protection depending on brightness level.

	# Items:
	Number    brightness                    "Current brightness [%d lux]"       {channel="..."}
	Number    brightness_threshold          "Brightness threshold [%d lux]"
	Switch    sun_protection_brightness     "Sun protection brightness [%s]     {channel="..."}

	# Rule init:
	habapp_rules.sensors.sun.SensorBrightness("brightness", "sun_protection_brightness", "brightness_threshold")
	"""

	def __init__(self, name_brightness: str, name_output: str, threshold: str | float, filter_tau: int = 20 * 60, filter_instant_increase: bool = True, filter_instant_decrease: bool = False) -> None:
		"""Init of sun sensor which takes a brightness value

		:param name_brightness: name of OpenHAB brightness item (NumberItem)
		:param name_output: name of OpenHAB output item (SwitchItem)
		:param threshold: threshold for the temperature difference which is supposed that sun is shining. Can be given as float value or name of OpenHAB NumberItem
		:param filter_tau: filter constant for the exponential filter. Default is set to 20 minutes
		:param filter_instant_increase: if set to True, increase of input values will not be filtered
		:param filter_instant_decrease: if set to True, decrease of input values will not be filtered
		"""
		_SensorBase.__init__(self, name_brightness, name_output, threshold, filter_tau, filter_instant_increase, filter_instant_decrease)


class SensorTemperatureDifference(_SensorBase):
	"""Rules class to set sun protection depending on temperature difference. E.g. temperature in the sun / temperature in the shadow.

	# Items:
	Number    temperature_sun               "Temperature sun [%.1f °C]"         {channel="..."}
	Number    temperature_shadow            "Temperature shadow [%.1f °C]"      {channel="..."}
	Number    temperature_threshold         "Temperature threshold [%.1f °C]"
	Switch    sun_protection_temperature    "Sun protection temperature [%s]    {channel="..."}

	# Rule init:
	habapp_rules.sensors.sun.SensorTempDiff("temperature_sun", "temperature_shadow", "sun_protection_temperature", "temperature_threshold")
	"""

	def __init__(self,
	             temperature_item_names: list[str],
	             name_output: str,
	             threshold: str | float,
	             filter_tau: int = 30 * 60,
	             filter_instant_increase: bool = True,
	             filter_instant_decrease: bool = False,
	             ignore_old_values_time: int | None = None) -> None:
		"""Init of sun sensor which takes a two or more temperature values (one in the sun and one in the shadow)

		:param temperature_item_names: name of all OpenHAB temperature items (NumberItem)
		:param name_output: name of OpenHAB output item (SwitchItem)
		:param threshold: threshold for the temperature difference which is supposed that sun is shining. Can be given as float value or name of OpenHAB NumberItem
		:param filter_tau: filter constant for the exponential filter. Default is set to 30 minutes
		:param filter_instant_increase: if set to True, increase of input values will not be filtered
		:param filter_instant_decrease: if set to True, decrease of input values will not be filtered
		:param ignore_old_values_time: ignores values which are older than the given time in seconds. If None, all values will be taken
		"""
		self._ignore_old_values_time = ignore_old_values_time
		name_temperature_diff = f"H_Temperature_diff_for_{name_output}"
		habapp_rules.core.helper.create_additional_item(name_temperature_diff, "Number", name_temperature_diff.replace("_", " "))

		_SensorBase.__init__(self, name_temperature_diff, name_output, threshold, filter_tau, filter_instant_increase, filter_instant_decrease)

		# init items
		self._temperature_items = [HABApp.openhab.items.NumberItem.get_item(name) for name in temperature_item_names]
		self._item_temp_diff = HABApp.openhab.items.NumberItem.get_item(name_temperature_diff)

		# callbacks
		for temperature_item in self._temperature_items:
			temperature_item.listen_event(self._cb_temperature, HABApp.openhab.events.ItemStateChangedEventFilter())

		# calculate temperature difference
		self._cb_temperature(None)

	def _cb_temperature(self, _: HABApp.openhab.events.ItemStateChangedEvent | None):
		"""Callback, which is triggered if a temperature value changed."""
		filtered_items = [itm for itm in habapp_rules.core.helper.filter_updated_items(self._temperature_items, self._ignore_old_values_time) if itm.value is not None]
		if len(filtered_items) < 2:
			return
		value_min = min(item.value for item in filtered_items)
		value_max = max(item.value for item in filtered_items)

		self._item_temp_diff.oh_send_command(value_max - value_min)


class SunPositionFilter(HABApp.Rule):
	"""Rules class to filter a switch state depending on the sun position. This can be used to only close the blinds of a window, if the sun hits the window

	# Items:
	Number    sun_azimuth           "Sun Azimuth [%.1f °]"              {channel="astro..."}
	Number    sun_elevation         "Sun Elevation [%.1f °]"            {channel="astro..."}
	Switch    sun_shining           "Sun is shining [%s]
	Switch    sun_hits_window       "Sun hits window [%s]

	# Rule init:
	position_window = habapp_rules.sensors.sun.SunPositionWindow(40, 120)
	habapp_rules.sensors.sun.SunPositionFilter(position_window, "sun_azimuth", "sun_elevation", "sun_shining", "sun_hits_window")
	"""

	def __init__(self, sun_position_window: SunPositionWindow | list[SunPositionWindow], name_azimuth: str, name_elevation: str, name_input: str, name_output: str):
		"""Init of sun position filter.

		:param sun_position_window: sun position window, where the sun hits the target
		:param name_azimuth: azimuth of the sun
		:param name_elevation: elevation of the sun
		:param name_input: name of OpenHAB input item (SwitchItem)
		:param name_output: name of OpenHAB output item (SwitchItem)
		"""
		# init HABApp Rule
		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, name_output)
		self._position_windows: list[SunPositionWindow] = sun_position_window if isinstance(sun_position_window, list) else [sun_position_window]

		# init items
		self._item_azimuth = HABApp.openhab.items.NumberItem.get_item(name_azimuth)
		self._item_elevation = HABApp.openhab.items.NumberItem.get_item(name_elevation)
		self._item_input = HABApp.openhab.items.SwitchItem.get_item(name_input)
		self._item_output = HABApp.openhab.items.SwitchItem.get_item(name_output)

		# callbacks
		self._item_azimuth.listen_event(self._update_output, HABApp.openhab.events.ItemStateChangedEventFilter())  # listen_event for elevation is not needed because elevation and azimuth is updated together
		self._item_input.listen_event(self._update_output, HABApp.openhab.events.ItemStateChangedEventFilter())

		self._update_output(None)

	def _sun_in_window(self, azimuth: float, elevation: float) -> bool:
		"""Check if the sun is in the 'sun window' where it hits the target.

		:param azimuth: azimuth of the sun
		:param elevation: elevation of the sun
		:return: True if the sun hits the target, else False
		"""
		sun_in_window = False

		if any(window.azimuth_min <= azimuth <= window.azimuth_max and window.elevation_min <= elevation <= window.elevation_max for window in self._position_windows):
			sun_in_window = True

		return sun_in_window

	def _update_output(self, _: HABApp.openhab.events.ItemStateChangedEvent | None) -> None:
		"""Callback, which is triggered if the sun position or input changed."""
		azimuth = self._item_azimuth.value
		elevation = self._item_elevation.value

		if azimuth is None or elevation is None:
			self._instance_logger.warning(f"Azimuth or elevation is None -> will set output to input. azimuth = {azimuth} | elevation = {elevation}")
			filter_output = self._item_input.value
		elif self._item_input.value in {"OFF", None}:
			filter_output = "OFF"
		else:
			filter_output = "ON" if self._sun_in_window(azimuth, elevation) else "OFF"

		if filter_output != self._item_output.value:
			self._item_output.oh_send_command(filter_output)



"""DWD rules."""
import dataclasses
import datetime
import logging
import re

import HABApp

import habapp_rules.actors.state_observer
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule

LOGGER = logging.getLogger(__name__)


@dataclasses.dataclass
class DwdItems:
	"""Dataclass for DWD items needed by DwdWindAlarm."""

	description: HABApp.openhab.items.StringItem
	warn_type: HABApp.openhab.items.StringItem
	severity: HABApp.openhab.items.StringItem
	start_time: HABApp.openhab.items.DatetimeItem
	end_time: HABApp.openhab.items.DatetimeItem

	SEVERITY_MAPPING = {
		"NULL": 0,
		"Minor": 1,
		"Moderate": 2,
		"Severe": 3,
		"Extreme": 4
	}

	@classmethod
	def from_prefix(cls, prefix: str) -> "DwdItems":
		"""Init DwdItems from prefix

		:param prefix: common prefix of all DWD items
		:return: Dataclass with all needed DWD items
		"""
		description = HABApp.openhab.items.StringItem.get_item(f"{prefix}_description")
		warn_type = HABApp.openhab.items.StringItem.get_item(f"{prefix}_type")
		severity = HABApp.openhab.items.StringItem.get_item(f"{prefix}_severity")
		start_time = HABApp.openhab.items.DatetimeItem.get_item(f"{prefix}_start_time")
		end_time = HABApp.openhab.items.DatetimeItem.get_item(f"{prefix}_end_time")

		return cls(description, warn_type, severity, start_time, end_time)

	@property
	def severity_as_int(self) -> int:
		"""Get severity as integer.

		:return: severity as integer value
		"""
		return self.SEVERITY_MAPPING.get(self.severity.value, 0)


# pylint: disable=no-member
class DwdWindAlarm(habapp_rules.core.state_machine_rule.StateMachineRule):
	"""Rule for setting wind alarm by DWD warnings.

	# Items:
	Switch      I26_99_wind_alarm               "Wind alarm"
	Switch      I26_99_wind_alarm_manual        "Wind alarm manual"

	String      I26_99_warning_1_severity       "Severity"              {channel="dwdunwetter:dwdwarnings:ingolstadt:severity1"}
	String      I26_99_warning_1_description    "Description [%s]"      {channel="dwdunwetter:dwdwarnings:ingolstadt:description1"}
	DateTime    I26_99_warning_1_start_time     "valid from [%s]"       {channel="dwdunwetter:dwdwarnings:ingolstadt:onset1"}
	DateTime    I26_99_warning_1_end_time       "valid till [%s]"       {channel="dwdunwetter:dwdwarnings:ingolstadt:expires1"}
	String      I26_99_warning_1_type           "Type [%s]"             {channel="dwdunwetter:dwdwarnings:ingolstadt:event1"}

	String      I26_99_warning_2_severity       "Severity"              {channel="dwdunwetter:dwdwarnings:ingolstadt:severity2"}
	String      I26_99_warning_2_description    "Description [%s]"      {channel="dwdunwetter:dwdwarnings:ingolstadt:description2"}
	DateTime    I26_99_warning_2_start_time     "valid from [%s]"       {channel="dwdunwetter:dwdwarnings:ingolstadt:onset2"}
	DateTime    I26_99_warning_2_end_time       "valid till [%s]"       {channel="dwdunwetter:dwdwarnings:ingolstadt:expires2"}
	String      I26_99_warning_2_type           "Type [%s]"             {channel="dwdunwetter:dwdwarnings:ingolstadt:event2"}

	# Rule init:
	habapp_rules.sensors.dwd.DwdWindAlarm(
			"I26_99_wind_alarm",
			"I26_99_wind_alarm_manual",
			12 * 3600,
			number_dwd_objects=2
	)
	"""

	states = [
		{"name": "Manual"},
		{"name": "Hand", "timeout": 20 * 3600, "on_timeout": "_auto_hand_timeout"},
		{"name": "Auto", "initial": "Init", "children": [
			{"name": "Init"},
			{"name": "On"},
			{"name": "Off"}
		]}
	]

	trans = [
		{"trigger": "manual_on", "source": ["Auto", "Hand"], "dest": "Manual"},
		{"trigger": "manual_off", "source": "Manual", "dest": "Auto"},
		{"trigger": "hand", "source": "Auto", "dest": "Hand"},

		{"trigger": "wind_alarm_start", "source": "Auto_Off", "dest": "Auto_On"},
		{"trigger": "wind_alarm_end", "source": "Auto_On", "dest": "Auto_Off"},
	]

	def __init__(self,
	             name_wind_alarm: str,
	             manual_name: str,
	             hand_timeout: str | int,
	             dwd_item_prefix: str = "I26_99_warning_",
	             number_dwd_objects: int = 3,
	             threshold_wind_speed: int = 70,
	             threshold_severity: int = 2,
	             name_state: str | None = None,
	             state_label: str | None = None) -> None:
		"""Init of DWD wind alarm object.

		:param name_wind_alarm: name of OpenHAB wind alarm item (SwitchItem)
		:param manual_name: name of OpenHAB switch item to disable all automatic functions (SwitchItem)
		:param hand_timeout: name of OpenHAB number item or fixed value as integer to set timeout of manual change
		:param dwd_item_prefix: [optional] prefix of dwd warning names
		:param number_dwd_objects: [optional] number of dwd objects
		:param threshold_wind_speed: [optional] threshold for wind speed -> wind alarm will only be active if greater or equal
		:param threshold_severity: [optional] threshold for severity -> wind alarm will only be active if greater or equal
		:param name_state: [optional] name of OpenHAB item for storing the current state (StringItem)
		:param state_label: [optional] label of OpenHAB item for storing the current state (StringItem)
		:raises TypeError: if type of hand_timeout is not supported
		"""
		if not name_state:
			name_state = f"H_{name_wind_alarm}_state"
		habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, name_state, state_label)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, name_wind_alarm)

		self._threshold_wind_speed = threshold_wind_speed
		self._threshold_severity = threshold_severity

		# init items
		self._item_manual = HABApp.openhab.items.switch_item.SwitchItem.get_item(manual_name)
		if isinstance(hand_timeout, str):
			self._item_hand_timeout = HABApp.openhab.items.number_item.NumberItem.get_item(hand_timeout)
			self._hand_timeout = None
		elif isinstance(hand_timeout, int):
			self._item_hand_timeout = None
			self._hand_timeout = hand_timeout
		else:
			raise TypeError(f"hand_timeout must be of type string or integer! Given value: {hand_timeout} | type = {type(hand_timeout)}")
		self._items_dwd = [DwdItems.from_prefix(f"{dwd_item_prefix}{idx + 1}") for idx in range(number_dwd_objects)]

		# init state machine
		self._previous_state = None
		self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(
			model=self,
			states=self.states,
			transitions=self.trans,
			ignore_invalid_triggers=True,
			after_state_change="_update_openhab_state")  # this function is missing!

		self._wind_alarm_observer = habapp_rules.actors.state_observer.StateObserverSwitch(name_wind_alarm, self._cb_hand, self._cb_hand)

		self._set_timeouts()
		self._set_initial_state()

		# callbacks
		self._item_manual.listen_event(self._cb_manual, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_hand_timeout is not None:
			self._item_hand_timeout.listen_event(self._cb_hand_timeout, HABApp.openhab.events.ItemStateChangedEventFilter())

		self.run.every(None, 300, self._cb_cyclic_check)

	def _get_initial_state(self, default_value: str = "") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: if OpenHAB item has a state it will return it, otherwise return the given default value
		"""
		if self._item_manual.is_on():
			return "Manual"
		if self._wind_alarm_active():
			return "Auto_On"
		return "Auto_Off"

	def _get_hand_timeout(self) -> int:
		"""Get value of hand timeout.

		:return: hand timeout in seconds
		"""
		if self._item_hand_timeout is not None:
			itm_value = self._item_hand_timeout.value
			return itm_value if itm_value is not None else 24 * 3600
		return self._hand_timeout

	def _set_timeouts(self) -> None:
		"""Set timeouts."""
		self.state_machine.get_state("Hand").timeout = self._get_hand_timeout()

	def _update_openhab_state(self) -> None:
		"""Update OpenHAB state item and other states.

		This should method should be set to "after_state_change" of the state machine.
		"""
		if self.state != self._previous_state:
			super()._update_openhab_state()
			self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

			if self.state == "Auto_On" and self._wind_alarm_observer.value is not True:
				self._wind_alarm_observer.send_command("ON")
			elif self.state == "Auto_Off" and self._wind_alarm_observer.value is not False:
				self._wind_alarm_observer.send_command("OFF")

			self._previous_state = self.state

	def on_enter_Auto_Init(self) -> None:  # pylint: disable=invalid-name
		"""Is called on entering of init state"""
		self._set_initial_state()

	def _cb_hand(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Callback, which is triggered by the state observer if a manual change was detected.

		:param event: original trigger event
		"""
		self.hand()

	def _cb_manual(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the manual switch has a state change event.

		:param event: trigger event
		"""
		if event.value == "ON":
			self.manual_on()
		else:
			self.manual_off()

	def _cb_hand_timeout(self, _) -> None:
		"""Callback which is triggered if the timeout item changed."""
		self._set_timeouts()

	def _wind_alarm_active(self) -> bool:
		"""Check if wind alarm is active

		:return: True if active, False if not
		"""
		for dwd_items in self._items_dwd:
			if dwd_items.warn_type.value in {"BÖEN", "WIND", "STURM", "GEWITTER"}:
				speed_values = [int(value) for value in re.findall(r"\b(\d+)\s*km/h\b", dwd_items.description.value)]
				if not speed_values:
					continue

				if max(speed_values) >= self._threshold_wind_speed and dwd_items.severity_as_int >= self._threshold_severity:
					if dwd_items.start_time.value < datetime.datetime.now() < dwd_items.end_time.value:
						return True
		return False

	def _cb_cyclic_check(self) -> None:
		"""Callback to check if wind alarm is active. This should be called cyclic every few minutes."""
		if not self.state in {"Auto_On", "Auto_Off"}:
			return

		if self._wind_alarm_active() and self.state != "Auto_On":
			self.wind_alarm_start()

		elif not self._wind_alarm_active() and self.state != "Auto_Off":
			self.wind_alarm_end()



"""Rules to manage lights."""
from __future__ import annotations

import abc
import copy
import logging
import math
import threading
import time
import typing

import HABApp.openhab.definitions
import HABApp.openhab.events
import HABApp.openhab.items
import HABApp.util

import habapp_rules.actors.config.light
import habapp_rules.actors.state_observer
import habapp_rules.core.exceptions
import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system

LOGGER = logging.getLogger(__name__)

BrightnessTypes = typing.Union[list[typing.Union[float, bool]], float, bool]


# pylint: disable=no-member, too-many-instance-attributes
class _LightBase(habapp_rules.core.state_machine_rule.StateMachineRule, metaclass=abc.ABCMeta):
	"""Base class for lights."""
	states = [
		{"name": "manual"},
		{"name": "auto", "initial": "init", "children": [
			{"name": "init"},
			{"name": "on", "timeout": 10, "on_timeout": "auto_on_timeout"},
			{"name": "preoff", "timeout": 4, "on_timeout": "preoff_timeout"},
			{"name": "off"},
			{"name": "leaving", "timeout": 5, "on_timeout": "leaving_timeout"},
			{"name": "presleep", "timeout": 5, "on_timeout": "presleep_timeout"},
			{"name": "restoreState"}
		]}
	]

	trans = [
		{"trigger": "manual_on", "source": "auto", "dest": "manual"},
		{"trigger": "manual_off", "source": "manual", "dest": "auto"},
		{"trigger": "hand_on", "source": ["auto_off", "auto_preoff"], "dest": "auto_on"},
		{"trigger": "hand_off", "source": ["auto_on", "auto_leaving", "auto_presleep"], "dest": "auto_off"},
		{"trigger": "hand_off", "source": "auto_preoff", "dest": "auto_on"},
		{"trigger": "auto_on_timeout", "source": "auto_on", "dest": "auto_preoff", "conditions": "_pre_off_configured"},
		{"trigger": "auto_on_timeout", "source": "auto_on", "dest": "auto_off", "unless": "_pre_off_configured"},
		{"trigger": "preoff_timeout", "source": "auto_preoff", "dest": "auto_off"},
		{"trigger": "leaving_started", "source": ["auto_on", "auto_off", "auto_preoff"], "dest": "auto_leaving", "conditions": "_leaving_configured"},
		{"trigger": "leaving_aborted", "source": "auto_leaving", "dest": "auto_restoreState"},
		{"trigger": "leaving_timeout", "source": "auto_leaving", "dest": "auto_off"},
		{"trigger": "sleep_started", "source": ["auto_on", "auto_off", "auto_preoff"], "dest": "auto_presleep", "conditions": "_pre_sleep_configured"},
		{"trigger": "sleep_aborted", "source": "auto_presleep", "dest": "auto_restoreState"},
		{"trigger": "presleep_timeout", "source": "auto_presleep", "dest": "auto_off"},
	]
	_item_light: HABApp.openhab.items.switch_item.SwitchItem | HABApp.openhab.items.dimmer_item.DimmerItem
	_state_observer: habapp_rules.actors.state_observer.StateObserverSwitch | habapp_rules.actors.state_observer.StateObserverDimmer

	def __init__(self,
	             name_light: str,
	             manual_name: str,
	             presence_state_name: str,
	             day_name: str,
	             config: habapp_rules.actors.config.light.LightConfig,
	             sleeping_state_name: str | None = None,
	             name_state: str | None = None,
	             state_label: str | None = None) -> None:
		"""Init of basic light object.

		:param name_light: name of OpenHAB light item (SwitchItem | DimmerItem)
		:param manual_name: name of OpenHAB switch item to disable all automatic functions
		:param presence_state_name: name of OpenHAB presence state item
		:param day_name: name of OpenHAB switch item which is 'ON' during day and 'OFF' during night
		:param config: configuration of the light object
		:param sleeping_state_name: [optional] name of OpenHAB sleeping state item
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		:raises TypeError: if type of light_item is not supported
		"""
		self._config = config

		if not name_state:
			name_state = f"H_{name_light}_state"
		habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, name_state, state_label)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, name_light)

		# init items
		self._item_manual = HABApp.openhab.items.switch_item.SwitchItem.get_item(manual_name)
		self._item_presence_state = HABApp.openhab.items.string_item.StringItem.get_item(presence_state_name)
		self._item_sleeping_state = HABApp.openhab.items.string_item.StringItem.get_item(sleeping_state_name) if sleeping_state_name else None
		self._item_day = HABApp.openhab.items.switch_item.SwitchItem.get_item(day_name)

		# init state machine
		self._previous_state = None
		self._restore_state = None
		self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(
			model=self,
			states=self.states,
			transitions=self.trans,
			ignore_invalid_triggers=True,
			after_state_change="_update_openhab_state")

		self._brightness_before = -1
		self._timeout_on = 0
		self._timeout_pre_off = 0
		self._timeout_pre_sleep = 0
		self._timeout_leaving = 0
		self.__time_sleep_start = 0
		self._set_timeouts()
		self._set_initial_state()

		# callbacks
		self._item_manual.listen_event(self._cb_manu, HABApp.openhab.events.ItemStateUpdatedEventFilter())
		if self._item_sleeping_state is not None:
			self._item_sleeping_state.listen_event(self._cb_sleeping, HABApp.openhab.events.ItemStateChangedEventFilter())
		self._item_presence_state.listen_event(self._cb_presence, HABApp.openhab.events.ItemStateChangedEventFilter())
		self._item_day.listen_event(self._cb_day, HABApp.openhab.events.ItemStateChangedEventFilter())

		self._update_openhab_state()
		self._instance_logger.debug(super().get_initial_log_message())

	def _get_initial_state(self, default_value: str = "") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: if OpenHAB item has a state it will return it, otherwise return the given default value
		"""
		if self._item_manual.is_on():
			return "manual"
		if self._item_light.is_on():
			if self._item_presence_state.value == habapp_rules.system.PresenceState.PRESENCE.value and \
					getattr(self._item_sleeping_state, "value", "awake") in (habapp_rules.system.SleepState.AWAKE.value, habapp_rules.system.SleepState.POST_SLEEPING.value, habapp_rules.system.SleepState.LOCKED.value):
				return "auto_on"
			if self._pre_sleep_configured() and \
					self._item_presence_state.value in (habapp_rules.system.PresenceState.PRESENCE.value, habapp_rules.system.PresenceState.LEAVING.value) and \
					getattr(self._item_sleeping_state, "value", "") in (habapp_rules.system.SleepState.PRE_SLEEPING.value, habapp_rules.system.SleepState.SLEEPING.value):
				return "auto_presleep"
			if self._leaving_configured():
				return "auto_leaving"
			return "auto_on"
		return "auto_off"

	def _update_openhab_state(self) -> None:
		"""Update OpenHAB state item and other states.

		This should method should be set to "after_state_change" of the state machine.
		"""
		if self.state != self._previous_state:
			super()._update_openhab_state()
			self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")

			self._set_light_state()
			self._previous_state = self.state

	def _pre_off_configured(self) -> bool:
		"""Check whether pre-off is configured for the current day/night/sleep state

		:return: True if pre-off is configured
		"""
		return bool(self._timeout_pre_off)

	def _leaving_configured(self) -> bool:
		"""Check whether leaving is configured for the current day/night/sleep state

		:return: True if leaving is configured
		"""
		return bool(self._timeout_leaving)

	def _pre_sleep_configured(self) -> bool:
		"""Check whether pre-sleep is configured for the current day/night state

		:return: True if pre-sleep is configured
		"""
		if self._item_sleeping_state is None:
			return False

		pre_sleep_prevent = False
		if self._config.pre_sleep_prevent:
			if callable(self._config.pre_sleep_prevent):
				pre_sleep_prevent = self._config.pre_sleep_prevent()
			if isinstance(self._config.pre_sleep_prevent, HABApp.openhab.items.OpenhabItem):
				pre_sleep_prevent = bool(self._config.pre_sleep_prevent)

		return bool(self._timeout_pre_sleep) and not pre_sleep_prevent

	def on_enter_auto_restoreState(self):  # pylint: disable=invalid-name
		"""On enter of state auto_restoreState."""
		self._restore_state = "auto_off" if self._restore_state == "auto_preoff" else self._restore_state

		if self._restore_state:
			self._set_state(self._restore_state)

	def _was_on_before(self) -> bool:
		"""Check whether the dimmer was on before

		:return: True if the dimmer was on before, else False
		"""
		return bool(self._brightness_before)

	def _set_timeouts(self) -> None:
		"""Set timeouts depending on the current day/night/sleep state."""
		if self._get_sleeping_activ():
			self._timeout_on = self._config.on.sleeping.timeout
			self._timeout_pre_off = getattr(self._config.pre_off.sleeping if self._config.pre_off else None, "timeout", None)
			self._timeout_leaving = getattr(self._config.leaving.sleeping if self._config.leaving else None, "timeout", None)
			self._timeout_pre_sleep = None

		elif self._item_day.is_on():
			self._timeout_on = self._config.on.day.timeout
			self._timeout_pre_off = getattr(self._config.pre_off.day if self._config.pre_off else None, "timeout", None)
			self._timeout_leaving = getattr(self._config.leaving.day if self._config.leaving else None, "timeout", None)
			self._timeout_pre_sleep = getattr(self._config.pre_sleep.day if self._config.pre_sleep else None, "timeout", None)
		else:
			self._timeout_on = self._config.on.night.timeout
			self._timeout_pre_off = getattr(self._config.pre_off.night if self._config.pre_off else None, "timeout", None)
			self._timeout_leaving = getattr(self._config.leaving.night if self._config.leaving else None, "timeout", None)
			self._timeout_pre_sleep = getattr(self._config.pre_sleep.night if self._config.pre_sleep else None, "timeout", None)

		self.state_machine.states["auto"].states["on"].timeout = self._timeout_on
		self.state_machine.states["auto"].states["preoff"].timeout = self._timeout_pre_off
		self.state_machine.states["auto"].states["leaving"].timeout = self._timeout_leaving
		self.state_machine.states["auto"].states["presleep"].timeout = self._timeout_pre_sleep

	@abc.abstractmethod
	def _set_light_state(self) -> None:
		"""Set brightness to light."""

	# pylint: disable=too-many-branches, too-many-return-statements
	def _get_target_brightness(self) -> bool | float | None:
		"""Get configured brightness for the current day/night/sleep state

		:return: brightness value
		"""
		sleeping_active = self._get_sleeping_activ(True)

		if self.state == "auto_on":
			if self._previous_state == "manual":
				return None
			if self._previous_state in {"auto_preoff", "auto_leaving", "auto_presleep"}:
				return self._brightness_before

			# starting from here: previous state == auto_off
			if isinstance(man_value := self._state_observer.last_manual_event.value, (int, float)) and 0 < man_value < 100:
				return None
			if self._state_observer.last_manual_event.value == "INCREASE":
				return None

			if sleeping_active:
				brightness_from_config = self._config.on.sleeping.brightness
			elif self._item_day.is_on():
				brightness_from_config = self._config.on.day.brightness
			else:
				brightness_from_config = self._config.on.night.brightness

			if brightness_from_config is True and self._state_observer.last_manual_event.value == "ON":
				return None

			return brightness_from_config

		if self.state == "auto_preoff":
			self._brightness_before = self._state_observer.value

			if sleeping_active:
				brightness_from_config = getattr(self._config.pre_off.sleeping if self._config.pre_off else None, "brightness", None)
			elif self._item_day.is_on():
				brightness_from_config = getattr(self._config.pre_off.day if self._config.pre_off else None, "brightness", None)
			else:
				brightness_from_config = getattr(self._config.pre_off.night if self._config.pre_off else None, "brightness", None)

			if brightness_from_config is None:
				return None

			if isinstance(self._state_observer.value, (float, int)) and brightness_from_config > self._state_observer.value:
				return math.ceil(self._state_observer.value / 2)
			return brightness_from_config

		if self.state == "auto_off":
			if self._previous_state == "manual":
				return None
			return False

		if self.state == "auto_presleep":
			if self._item_day.is_on():
				return getattr(self._config.pre_sleep.day if self._config.pre_sleep else None, "brightness", None)
			return getattr(self._config.pre_sleep.night if self._config.pre_sleep else None, "brightness", None)

		if self.state == "auto_leaving":
			if sleeping_active:
				return getattr(self._config.leaving.sleeping if self._config.leaving else None, "brightness", None)
			if self._item_day.is_on():
				return getattr(self._config.leaving.day if self._config.leaving else None, "brightness", None)
			return getattr(self._config.leaving.night if self._config.leaving else None, "brightness", None)

		return None

	def on_enter_auto_init(self):
		"""Callback, which is called on enter of init state"""
		if self._item_light.is_on():
			self.to_auto_on()
		else:
			self.to_auto_off()

	def _get_sleeping_activ(self, include_pre_sleep: bool = False) -> bool:
		"""Get if sleeping is active.

		:param include_pre_sleep: if true, also pre sleep will be handled as sleeping
		:return: true if sleeping active
		"""
		sleep_states = [habapp_rules.system.SleepState.PRE_SLEEPING.value, habapp_rules.system.SleepState.SLEEPING.value] if include_pre_sleep else [habapp_rules.system.SleepState.SLEEPING.value]
		return getattr(self._item_sleeping_state, "value", "") in sleep_states

	def _cb_hand_on(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Callback, which is triggered by the state observer if a manual ON command was detected.

		:param event: original trigger event
		"""
		self._instance_logger.debug("Hand 'ON' detected")
		self.hand_on()

	def _cb_hand_off(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Callback, which is triggered by the state observer if a manual OFF command was detected.

		:param event: original trigger event
		"""
		self._instance_logger.debug("Hand 'OFF' detected")
		self.hand_off()

	def _cb_manu(self, event: HABApp.openhab.events.ItemStateUpdatedEvent) -> None:
		"""Callback, which is triggered if the manual switch has a state event.

		:param event: trigger event
		"""
		if event.value == "ON":
			self.manual_on()
		else:
			self.manual_off()

	def _cb_day(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the day/night switch has a state change event.

		:param event: trigger event
		"""
		self._set_timeouts()

	def _cb_presence(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the presence state has a state change event.

		:param event: trigger event
		"""
		self._set_timeouts()
		if event.value == habapp_rules.system.PresenceState.LEAVING.value:
			self._brightness_before = self._state_observer.value
			self._restore_state = self._previous_state
			self.leaving_started()
		elif event.value == habapp_rules.system.PresenceState.PRESENCE.value and self.state == "auto_leaving":
			self.leaving_aborted()

	def _cb_sleeping(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the sleep state has a state change event.

		:param event: trigger event
		"""
		self._set_timeouts()
		if event.value == habapp_rules.system.SleepState.PRE_SLEEPING.value:
			self._brightness_before = self._state_observer.value
			self._restore_state = self._previous_state
			self.__time_sleep_start = time.time()
			self.sleep_started()
		elif event.value == habapp_rules.system.SleepState.AWAKE.value and time.time() - self.__time_sleep_start <= 60:
			self.sleep_aborted()


class LightSwitch(_LightBase):
	"""Rules class to manage basic light states.

		# KNX-things:
		Thing device T00_99_OpenHab_DimmerObserver "KNX OpenHAB dimmer observer"{
			Type switch             : light             "Light"             [ switch="1/1/10+1/1/13" ]
		}

		# Items:
		Switch    I01_01_Light              "Light [%s]"        {channel="knx:device:bridge:T00_99_OpenHab_DimmerObserver:light"}
		Switch    I00_00_Light_manual       "Light manual"

		# Rule init:
		habapp_rules.actors.light.LightSwitch(
			"I01_01_Light",
			manual_name="I00_00_Light_manual",
			presence_state_name="I999_00_Presence_state", # string item!
			sleeping_state_name="I999_00_Sleeping_state", # string item!
			day_name="I999_00_Day",
			config=CONFIG_TEST,
		)
		"""

	def __init__(self,
	             name_light: str,
	             manual_name: str,
	             presence_state_name: str,
	             day_name: str,
	             config: habapp_rules.actors.config.light.LightConfig,
	             sleeping_state_name: str | None = None,
	             name_state: str | None = None,
	             state_label: str | None = None):
		"""Init of basic light object.

		:param name_light: name of OpenHAB light item (SwitchItem)
		:param manual_name: name of OpenHAB switch item to disable all automatic functions
		:param presence_state_name: name of OpenHAB presence state item
		:param day_name: name of OpenHAB switch item which is 'ON' during day and 'OFF' during night
		:param config: configuration of the light object
		:param sleeping_state_name: [optional] name of OpenHAB sleeping state item
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		:raises TypeError: if type of light_item is not supported
		"""
		light_item = HABApp.core.Items.get_item(name_light)
		if not isinstance(light_item, HABApp.openhab.items.switch_item.SwitchItem):
			raise TypeError(f"type: {type(light_item)} is not supported!")

		self._item_light = HABApp.openhab.items.SwitchItem.get_item(name_light)
		self._state_observer = habapp_rules.actors.state_observer.StateObserverSwitch(name_light, self._cb_hand_on, self._cb_hand_off)

		_LightBase.__init__(self, name_light, manual_name, presence_state_name, day_name, config, sleeping_state_name, name_state, state_label)

	def _update_openhab_state(self) -> None:
		_LightBase._update_openhab_state(self)

		if self.state == "auto_preoff":
			timeout = self.state_machine.get_state(self.state).timeout

			warn_thread_1 = threading.Thread(target=self.__trigger_warning, args=("auto_preoff", 0, 1), daemon=True)
			warn_thread_1.start()

			if timeout > 60:
				# add additional warning for long timeouts
				warn_thread_2 = threading.Thread(target=self.__trigger_warning, args=("auto_preoff", timeout / 2, 2), daemon=True)
				warn_thread_2.start()

	def __trigger_warning(self, state_name: str, wait_time: float, switch_off_amount: int) -> None:
		"""Trigger light switch off warning.

		:param state_name: name of state where the warning should be triggered. If different no command will be sent
		:param wait_time: time between start of the thread and switch off / on
		:param switch_off_amount: number of switch off
		"""

		if wait_time:
			time.sleep(wait_time)

		for idx in range(switch_off_amount):
			if self.state != state_name:
				break
			self._state_observer.send_command("OFF")
			time.sleep(0.2)
			if self.state != state_name:
				break
			self._state_observer.send_command("ON")
			if idx + 1 < switch_off_amount:
				time.sleep(0.5)

	def _set_light_state(self) -> None:
		"""Set brightness to light."""
		target_value = self._get_target_brightness()
		if target_value is None or self._previous_state is None:
			# don't change value if target_value is None or _set_light_state will be called during init (_previous_state == None)
			return

		target_value = "ON" if target_value else "OFF"
		self._instance_logger.debug(f"set brightness {target_value}")
		self._state_observer.send_command(target_value)


class LightDimmer(_LightBase):
	"""Rules class to manage basic light states.

	# KNX-things:
	Thing device T00_99_OpenHab_DimmerObserver "KNX OpenHAB dimmer observer"{
		Type dimmer             : light             "Light"             [ switch="1/1/10", position="1/1/13+<1/1/15" ]
		Type dimmer-control     : light_ctr         "Light control"     [ increaseDecrease="1/1/12"]
		Type dimmer             : light_group       "Light Group"       [ switch="1/1/240", position="1/1/243"]
	}

	# Items:
	Dimmer    I01_01_Light              "Light [%s]"        {channel="knx:device:bridge:T00_99_OpenHab_DimmerObserver:light"}
	Dimmer    I01_01_Light_ctr          "Light ctr"         {channel="knx:device:bridge:T00_99_OpenHab_DimmerObserver:light_ctr"}
	Dimmer    I01_01_Light_group        "Light Group"       {channel="knx:device:bridge:T00_99_OpenHab_DimmerObserver:light_group"}
	Switch    I00_00_Light_manual       "Light manual"

	# Rule init:
	habapp_rules.actors.light.LightDimmer(
		"I01_01_Light",
		control_names=["I01_01_Light_ctr"],
		manual_name="I00_00_Light_manual",
		presence_state_name="I999_00_Presence_state", # string item!
		sleeping_state_name="I999_00_Sleeping_state", # string item!
		day_name="I999_00_Day",
		config=CONFIG_TEST,
		group_names=["I01_01_Light_group"]
	)
	"""

	trans = copy.deepcopy(_LightBase.trans)
	trans.append({"trigger": "hand_changed", "source": "auto_on", "dest": "auto_on"})

	# pylint: disable=too-many-arguments
	def __init__(self,
	             name_light: str,
	             control_names: list[str],
	             manual_name: str,
	             presence_state_name: str,
	             day_name: str,
	             config: habapp_rules.actors.config.light.LightConfig,
	             sleeping_state_name: str | None = None,
	             name_state: str | None = None,
	             state_label: str | None = None,
	             group_names: list[str] | None = None) -> None:
		"""Init of basic light object.

		:param name_light: name of OpenHAB light item (DimmerItem)
		:param control_names: names of OpenHab items which must be configured as control (-ctr) items. This can be used for KNX items to detect increase / decrease commands from physical wall controllers
		:param manual_name: name of OpenHAB switch item to disable all automatic functions
		:param presence_state_name: name of OpenHAB presence state item
		:param day_name: name of OpenHAB switch item which is 'ON' during day and 'OFF' during night
		:param config: configuration of the light object
		:param sleeping_state_name: [optional] name of OpenHAB sleeping state item
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		:param group_names: list of group items where the light is a part of. Group item type must match with type of the light item
		:raises TypeError: if type of light_item is not supported
		"""
		light_item = HABApp.core.Items.get_item(name_light)
		if not isinstance(light_item, HABApp.openhab.items.dimmer_item.DimmerItem):
			raise TypeError(f"type: {type(light_item)} is not supported!")

		self._item_light = HABApp.openhab.items.DimmerItem.get_item(name_light)
		self._state_observer = habapp_rules.actors.state_observer.StateObserverDimmer(name_light, self._cb_hand_on, self._cb_hand_off, self._cb_hand_changed, control_names=control_names, group_names=group_names)

		_LightBase.__init__(self, name_light, manual_name, presence_state_name, day_name, config, sleeping_state_name, name_state, state_label)

	def _set_light_state(self) -> None:
		"""Set brightness to light."""
		target_value = self._get_target_brightness()
		if target_value is None or self._previous_state is None:
			# don't change value if target_value is None or _set_light_state will be called during init (_previous_state == None)
			return

		if isinstance(target_value, bool):
			if target_value:
				target_value = "ON"
			else:
				target_value = "OFF"
		self._instance_logger.debug(f"set brightness {target_value}")
		self._state_observer.send_command(target_value)

	def _cb_hand_changed(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | HABApp.openhab.events.ItemCommandEvent | HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered by the state observer if a manual OFF command was detected.

		:param event: original trigger event
		"""
		if isinstance(event, HABApp.openhab.events.ItemStateChangedEvent) and abs(event.value - event.old_value) > 5:
			self.hand_changed()


# pylint: disable=protected-access
class _LightExtendedMixin:
	"""Mixin class for adding door and motion functionality"""
	states: dict
	trans: list
	state: str
	state_machine: habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout
	_config: habapp_rules.actors.config.light.LightConfigExtended
	_get_sleeping_activ: typing.Callable[[bool | None], bool]
	_item_day: HABApp.openhab.items.SwitchItem

	def __init__(self, config: habapp_rules.actors.config.light.LightConfigExtended, name_motion: str | None = None, door_names: list[str] | None = None):
		"""Init mixin class.

		:param config: config of light object
		:param name_motion: [optional] name of OpenHAB motion item (SwitchItem)
		:param door_names: [optional] list of OpenHAB door items (ContactItem)
		"""
		self.states = _LightExtendedMixin._add_additional_states(self.states)
		self.trans = _LightExtendedMixin._add_additional_transitions(self.trans)

		door_names = door_names if door_names else []

		self._timeout_motion = 0
		self._timeout_door = 0

		self._item_motion = HABApp.openhab.items.switch_item.SwitchItem.get_item(name_motion) if name_motion else None
		self._items_door = [HABApp.openhab.items.contact_item.ContactItem.get_item(name) for name in door_names]

		self._hand_off_lock_time = config.hand_off_lock_time
		self._hand_off_timestamp = 0

	@staticmethod
	def _add_additional_states(states_dict: dict) -> dict:
		"""Add additional states for door and motion.

		:param states_dict: current state dictionary
		:return: current + new states
		"""
		states_dict = copy.deepcopy(states_dict)
		states_dict[1]["children"].append({"name": "door", "timeout": 999, "on_timeout": "door_timeout"})
		states_dict[1]["children"].append({"name": "motion", "timeout": 999, "on_timeout": "motion_timeout"})
		return states_dict

	@staticmethod
	def _add_additional_transitions(transitions_list: list[dict]) -> list[dict]:
		"""Add additional transitions for door and motion

		:param transitions_list: current transitions
		:return: current + new transitions
		"""
		transitions_list = copy.deepcopy(transitions_list)

		transitions_list.append({"trigger": "motion_on", "source": "auto_door", "dest": "auto_motion", "conditions": "_motion_configured"})
		transitions_list.append({"trigger": "motion_on", "source": "auto_off", "dest": "auto_motion", "conditions": ["_motion_configured", "_motion_door_allowed"]})
		transitions_list.append({"trigger": "motion_on", "source": "auto_preoff", "dest": "auto_motion", "conditions": "_motion_configured"})
		transitions_list.append({"trigger": "motion_off", "source": "auto_motion", "dest": "auto_preoff", "conditions": "_pre_off_configured"})
		transitions_list.append({"trigger": "motion_off", "source": "auto_motion", "dest": "auto_off", "unless": "_pre_off_configured"})
		transitions_list.append({"trigger": "motion_timeout", "source": "auto_motion", "dest": "auto_preoff", "conditions": "_pre_off_configured", "before": "_log_motion_timeout_warning"})
		transitions_list.append({"trigger": "motion_timeout", "source": "auto_motion", "dest": "auto_off", "unless": "_pre_off_configured", "before": "_log_motion_timeout_warning"})
		transitions_list.append({"trigger": "hand_off", "source": "auto_motion", "dest": "auto_off"})

		transitions_list.append({"trigger": "door_opened", "source": ["auto_off", "auto_preoff"], "dest": "auto_door", "conditions": ["_door_configured", "_motion_door_allowed"]})
		transitions_list.append({"trigger": "door_timeout", "source": "auto_door", "dest": "auto_preoff", "conditions": "_pre_off_configured"})
		transitions_list.append({"trigger": "door_timeout", "source": "auto_door", "dest": "auto_off", "unless": "_pre_off_configured"})
		transitions_list.append({"trigger": "door_closed", "source": "auto_leaving", "dest": "auto_off", "conditions": "_door_off_leaving_configured"})
		transitions_list.append({"trigger": "hand_off", "source": "auto_door", "dest": "auto_off"})

		transitions_list.append({"trigger": "leaving_started", "source": ["auto_motion", "auto_door"], "dest": "auto_leaving", "conditions": "_leaving_configured"})
		transitions_list.append({"trigger": "sleep_started", "source": ["auto_motion", "auto_door"], "dest": "auto_presleep", "conditions": "_pre_sleep_configured"})

		return transitions_list

	def add_additional_callbacks(self) -> None:
		"""Add additional callbacks for motion and door items."""
		if self._item_motion is not None:
			self._item_motion.listen_event(self._cb_motion, HABApp.openhab.events.ItemStateChangedEventFilter())
		for item_door in self._items_door:
			item_door.listen_event(self._cb_door, HABApp.openhab.events.ItemStateChangedEventFilter())

	def _get_initial_state(self, default_value: str = "") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: if OpenHAB item has a state it will return it, otherwise return the given default value
		"""
		initial_state = _LightBase._get_initial_state(self, default_value)

		if initial_state == "auto_on" and self._item_motion is not None and self._item_motion.is_on() and self._motion_configured():
			initial_state = "auto_motion"
		return initial_state

	def _set_timeouts(self) -> None:
		"""Set timeouts depending on the current day/night/sleep state."""
		_LightBase._set_timeouts(self)

		# set timeouts of additional states
		if self._get_sleeping_activ():
			self._timeout_motion = getattr(self._config.motion.sleeping if self._config.motion else None, "timeout", None)
			self._timeout_door = getattr(self._config.door.sleeping if self._config.door else None, "timeout", None)

		elif self._item_day.is_on():
			self._timeout_motion = getattr(self._config.motion.day if self._config.motion else None, "timeout", None)
			self._timeout_door = getattr(self._config.door.day if self._config.door else None, "timeout", None)
		else:
			self._timeout_motion = getattr(self._config.motion.night if self._config.motion else None, "timeout", None)
			self._timeout_door = getattr(self._config.door.night if self._config.door else None, "timeout", None)

		self.state_machine.states["auto"].states["motion"].timeout = self._timeout_motion
		self.state_machine.states["auto"].states["door"].timeout = self._timeout_door

	def _get_target_brightness(self) -> bool | float | None:
		"""Get configured brightness for the current day/night/sleep state. Must be called before _get_target_brightness of base class

		:return: brightness value
		:raises habapp_rules.core.exceptions.HabAppRulesException: if current state is not supported
		"""
		if self.state == "auto_motion":
			if self._get_sleeping_activ(True):
				return getattr(self._config.motion.sleeping if self._config.motion else None, "brightness", None)
			if self._item_day.is_on():
				return getattr(self._config.motion.day if self._config.motion else None, "brightness", None)
			return getattr(self._config.motion.night if self._config.motion else None, "brightness", None)

		if self.state == "auto_door":
			if self._get_sleeping_activ(True):
				return getattr(self._config.door.sleeping if self._config.door else None, "brightness", None)
			if self._item_day.is_on():
				return getattr(self._config.door.day if self._config.door else None, "brightness", None)
			return getattr(self._config.door.night if self._config.door else None, "brightness", None)

		return _LightBase._get_target_brightness(self)

	def _door_configured(self) -> bool:
		"""Check whether door functionality is configured for the current day/night state

		:return: True if door functionality is configured
		"""
		if not self._items_door:
			return False
		return bool(self._timeout_door)

	def _door_off_leaving_configured(self) -> bool:
		"""Check whether door-off functionality is configured for the current day/night state

		:return: True if door-off is configured
		"""
		return self._config.off_at_door_closed_during_leaving

	def _motion_configured(self) -> bool:
		"""Check whether motion functionality is configured for the current day/night state

		:return: True if motion functionality is configured
		"""
		if self._item_motion is None:
			return False
		return bool(self._timeout_motion)

	def _motion_door_allowed(self) -> bool:
		"""Check if transition to motion and door state is allowed

		:return: True if transition is allowed
		"""
		return time.time() - self._hand_off_timestamp > self._hand_off_lock_time

	def _cb_hand_off(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | HABApp.openhab.events.ItemCommandEvent) -> None:
		"""Callback, which is triggered by the state observer if a manual OFF command was detected.

		:param event: original trigger event
		"""
		self._hand_off_timestamp = time.time()
		_LightBase._cb_hand_off(self, event)

	def _cb_motion(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the motion state changed.

		:param event: trigger event
		"""
		if event.value == "ON":
			self.motion_on()
		else:
			self.motion_off()

	def _cb_door(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if a door state changed.

		:param event: trigger event
		"""
		if event.value == "OPEN":
			# every open of a single door calls door_opened()
			self.door_opened()

		if event.value == "CLOSED" and all(door.is_closed() for door in self._items_door):
			# only if all doors are closed door_closed() is called
			self.door_closed()

	def _log_motion_timeout_warning(self):
		"""Log warning if motion state was left because of timeout."""
		self._instance_logger.warning("Timeout of motion was triggered, before motion stopped. Thing about to increase motion timeout!")


# pylint: disable=protected-access
class LightSwitchExtended(_LightExtendedMixin, LightSwitch):
	"""Extended Light.

	Example config is given at Light base class.
	With this class additionally motion or door items can be given.
	"""

	# pylint: disable=too-many-arguments
	def __init__(self, name_light: str,
	             manual_name: str,
	             presence_state_name: str,
	             day_name: str,
	             config: habapp_rules.actors.config.light.LightConfigExtended, sleeping_state_name: str | None = None,
	             name_motion: str | None = None,
	             door_names: list[str] | None = None,
	             name_state: str | None = None,
	             state_label: str | None = None):
		"""Init of extended light object.

		:param name_light: name of OpenHAB light item (SwitchItem)
		:param manual_name: name of OpenHAB switch item to disable all automatic functions
		:param presence_state_name: name of OpenHAB presence state item
		:param day_name: name of OpenHAB switch item which is 'ON' during day and 'OFF' during night
		:param config: configuration of the light object
		:param sleeping_state_name: [optional] name of OpenHAB sleeping state item
		:param name_motion: [optional] name of OpenHAB motion item (SwitchItem)
		:param door_names: [optional] list of OpenHAB door items (ContactItem)
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		:raises TypeError: if type of light_item is not supported
		"""
		_LightExtendedMixin.__init__(self, config, name_motion, door_names)
		LightSwitch.__init__(self, name_light, manual_name, presence_state_name, day_name, config, sleeping_state_name, name_state, state_label)

		_LightExtendedMixin.add_additional_callbacks(self)


# pylint: disable=protected-access
class LightDimmerExtended(_LightExtendedMixin, LightDimmer):
	"""Extended Light.

	Example config is given at Light base class.
	With this class additionally motion or door items can be given.
	"""

	# pylint:disable=too-many-arguments
	def __init__(self, name_light: str,
	             control_names: list[str],
	             manual_name: str,
	             presence_state_name: str,
	             day_name: str,
	             config: habapp_rules.actors.config.light.LightConfigExtended,
	             sleeping_state_name: str | None = None,
	             name_motion: str | None = None,
	             door_names: list[str] | None = None,
	             name_state: str | None = None,
	             state_label: str | None = None,
	             group_names: list[str] | None = None) -> None:
		"""Init of extended light object.

		:param name_light: name of OpenHAB light item (DimmerItem)
		:param control_names: names of OpenHab items which must be configured as control (-ctr) items. This can be used for KNX items to detect increase / decrease commands from physical wall controllers
		:param manual_name: name of OpenHAB switch item to disable all automatic functions
		:param presence_state_name: name of OpenHAB presence state item
		:param day_name: name of OpenHAB switch item which is 'ON' during day and 'OFF' during night
		:param config: configuration of the light object
		:param sleeping_state_name: [optional] name of OpenHAB sleeping state item
		:param name_motion: [optional] name of OpenHAB motion item (SwitchItem)
		:param door_names: [optional] list of OpenHAB door items (ContactItem)
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		:param group_names: list of group items where the light is a part of. Group item type must match with type of the light item
		:raises TypeError: if type of light_item is not supported
		"""
		_LightExtendedMixin.__init__(self, config, name_motion, door_names)
		LightDimmer.__init__(self, name_light, control_names, manual_name, presence_state_name, day_name, config, sleeping_state_name, name_state, state_label, group_names)

		_LightExtendedMixin.add_additional_callbacks(self)



"""Rules for managing motion sensors."""
import logging

import HABApp

import habapp_rules.common.hysteresis
import habapp_rules.core.exceptions
import habapp_rules.core.helper
import habapp_rules.core.logger
import habapp_rules.core.state_machine_rule
import habapp_rules.system.sleep

LOGGER = logging.getLogger(__name__)


# pylint: disable=no-member, too-many-instance-attributes
class Motion(habapp_rules.core.state_machine_rule.StateMachineRule):
	"""Class for filtering motion sensors.

	# MQTT-things:
	Thing topic Motion "Motion Sensor"{
        Channels:
        Type switch : motion        "Motion"        [stateTopic="zigbee2mqtt/Motion/occupancy", on="true", off="false"]
        Type number : brightness    "Brightness"    [stateTopic="zigbee2mqtt/Motion/illuminance_lux"]
    }

    # Items:
    Switch    Motion_raw                "Motion raw"                <motion>        {channel="mqtt:topic:broker:Motion:motion"}
	Switch    Motion_filtered           "Motion filtered"           <motion>
	Number    Motion_Brightness         "Brightness"                                {channel="mqtt:topic:broker:Motion:brightness"}

	# Rule init:
	habapp_rules.sensors.motion.Motion(
		"Motion_raw",
		"Motion_filtered",
		name_brightness="Motion_Brightness",
		brightness_threshold=100,
		name_sleep_state="I999_00_Sleeping_state"
	)
	"""
	states = [
		{"name": "Locked"},
		{"name": "SleepLocked"},
		{"name": "PostSleepLocked", "timeout": 99, "on_timeout": "timeout_post_sleep_locked"},
		{"name": "Unlocked", "initial": "Init", "children": [
			{"name": "Init"},
			{"name": "Wait"},
			{"name": "Motion"},
			{"name": "MotionExtended", "timeout": 99, "on_timeout": "timeout_motion_extended"},
			{"name": "TooBright"},
		]}
	]

	trans = [
		# lock
		{"trigger": "lock_on", "source": ["Unlocked", "SleepLocked", "PostSleepLocked"], "dest": "Locked"},
		{"trigger": "lock_off", "source": "Locked", "dest": "Unlocked", "unless": "_sleep_active"},
		{"trigger": "lock_off", "source": "Locked", "dest": "SleepLocked", "conditions": "_sleep_active"},

		# sleep
		{"trigger": "sleep_started", "source": ["Unlocked", "PostSleepLocked"], "dest": "SleepLocked"},
		{"trigger": "sleep_end", "source": "SleepLocked", "dest": "Unlocked", "unless": "_post_sleep_lock_configured"},
		{"trigger": "sleep_end", "source": "SleepLocked", "dest": "PostSleepLocked", "conditions": "_post_sleep_lock_configured"},
		{"trigger": "timeout_post_sleep_locked", "source": "PostSleepLocked", "dest": "Unlocked", "unless": "_raw_motion_active"},
		{"trigger": "motion_off", "source": "PostSleepLocked", "dest": "PostSleepLocked"},
		{"trigger": "motion_on", "source": "PostSleepLocked", "dest": "PostSleepLocked"},

		# motion
		{"trigger": "motion_on", "source": "Unlocked_Wait", "dest": "Unlocked_Motion"},
		{"trigger": "motion_off", "source": "Unlocked_Motion", "dest": "Unlocked_MotionExtended", "conditions": "_motion_extended_configured"},
		{"trigger": "motion_off", "source": "Unlocked_Motion", "dest": "Unlocked_Wait", "unless": "_motion_extended_configured"},
		{"trigger": "timeout_motion_extended", "source": "Unlocked_MotionExtended", "dest": "Unlocked_Wait", "unless": "_brightness_over_threshold"},
		{"trigger": "timeout_motion_extended", "source": "Unlocked_MotionExtended", "dest": "Unlocked_TooBright", "conditions": "_brightness_over_threshold"},
		{"trigger": "motion_on", "source": "Unlocked_MotionExtended", "dest": "Unlocked_Motion"},

		# brightness
		{"trigger": "brightness_over_threshold", "source": "Unlocked_Wait", "dest": "Unlocked_TooBright"},
		{"trigger": "brightness_below_threshold", "source": "Unlocked_TooBright", "dest": "Unlocked_Wait", "unless": "_raw_motion_active"},
		{"trigger": "brightness_below_threshold", "source": "Unlocked_TooBright", "dest": "Unlocked_Motion", "conditions": "_raw_motion_active"}
	]

	# pylint: disable=too-many-arguments
	def __init__(self,
	             name_raw: str,
	             name_filtered: str,
	             extended_motion_time: int = 5,
	             name_brightness: str | None = None,
	             brightness_threshold: int | str | None = None,
	             name_lock: str | None = None,
	             name_sleep_state: str | None = None,
	             post_sleep_lock_time: int = 10,
	             name_state: str | None = None,
	             state_label: str | None = None) -> None:
		"""Init of motion filter.

		:param name_raw: name of OpenHAB unfiltered motion item (SwitchItem)
		:param name_filtered: name of OpenHAB filtered motion item (SwitchItem)
		:param extended_motion_time: time in seconds which will extend the motion after motion is off. If it is set to 0 the time will not be extended
		:param name_brightness: name of OpenHAB brightness item (NumberItem)
		:param brightness_threshold: brightness threshold value (float) or name of OpenHAB brightness threshold item (NumberItem)
		:param name_lock: name of OpenHAB lock item (SwitchItem)
		:param name_sleep_state: name of OpenHAB sleep state item (StringItem)
		:param post_sleep_lock_time: Lock time after sleep
		:param name_state: name of OpenHAB item for storing the current state (StringItem)
		:param state_label: label of OpenHAB item for storing the current state (StringItem)
		:raises habapp_rules.core.exceptions.HabAppRulesConfigurationException: if configuration is not valid
		"""
		if bool(name_brightness) != bool(brightness_threshold):
			raise habapp_rules.core.exceptions.HabAppRulesConfigurationException("'name_brightness' or 'brightness_threshold' is missing!")

		if not name_state:
			name_state = f"H_Motion_{name_raw}_state"

		habapp_rules.core.state_machine_rule.StateMachineRule.__init__(self, name_state, state_label)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, name_raw)
		self._brightness_threshold_value = brightness_threshold if isinstance(brightness_threshold, int) else None
		self._timeout_extended_motion = extended_motion_time
		self._timeout_post_sleep_lock = post_sleep_lock_time

		# get items
		self._item_motion_raw = HABApp.openhab.items.SwitchItem.get_item(name_raw)
		self._item_motion_filtered = HABApp.openhab.items.SwitchItem.get_item(name_filtered)
		self._item_brightness = HABApp.openhab.items.NumberItem.get_item(name_brightness) if name_brightness else None
		self._item_brightness_threshold = HABApp.openhab.items.NumberItem.get_item(brightness_threshold) if isinstance(brightness_threshold, str) else None
		self._item_lock = HABApp.openhab.items.SwitchItem.get_item(name_lock) if name_lock else None
		self._item_sleep = HABApp.openhab.items.StringItem.get_item(name_sleep_state) if name_sleep_state else None

		self._hysteresis_switch = habapp_rules.common.hysteresis.HysteresisSwitch(threshold := self._get_brightness_threshold(), threshold * 0.1 if threshold else 5) if name_brightness else None

		# init state machine
		self._previous_state = None
		self.state_machine = habapp_rules.core.state_machine_rule.HierarchicalStateMachineWithTimeout(
			model=self,
			states=self.states,
			transitions=self.trans,
			ignore_invalid_triggers=True,
			after_state_change="_update_openhab_state")
		self._set_initial_state()

		self.state_machine.get_state("PostSleepLocked").timeout = self._timeout_post_sleep_lock
		self.state_machine.get_state("Unlocked_MotionExtended").timeout = self._timeout_extended_motion

		# register callbacks
		self._item_motion_raw.listen_event(self._cb_motion_raw, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_brightness is not None:
			self._item_brightness.listen_event(self._cb_brightness, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_brightness_threshold is not None:
			self._item_brightness_threshold.listen_event(self._cb_threshold_change, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_lock is not None:
			self._item_lock.listen_event(self._cb_lock, HABApp.openhab.events.ItemStateChangedEventFilter())
		if self._item_sleep is not None:
			self._item_sleep.listen_event(self._cb_sleep, HABApp.openhab.events.ItemStateChangedEventFilter())

		self._instance_logger.debug(super().get_initial_log_message())

	def _get_initial_state(self, default_value: str = "initial") -> str:
		"""Get initial state of state machine.

		:param default_value: default / initial state
		:return: if OpenHAB item has a state it will return it, otherwise return the given default value
		"""
		if self._item_lock is not None and self._item_lock.is_on():
			return "Locked"
		if self._item_sleep is not None and self._item_sleep.value == habapp_rules.system.SleepState.SLEEPING.value:
			return "SleepLocked"
		if self._item_brightness is not None and self._brightness_over_threshold():
			return "Unlocked_TooBright"
		if self._item_motion_raw.is_on():
			return "Unlocked_Motion"
		return "Unlocked_Wait"

	def _update_openhab_state(self):
		"""Update OpenHAB state item. This should method should be set to "after_state_change" of the state machine."""
		if self.state != self._previous_state:
			super()._update_openhab_state()
			self.__send_filtered_motion()

			self._instance_logger.debug(f"State change: {self._previous_state} -> {self.state}")
			self._previous_state = self.state

	def __send_filtered_motion(self) -> None:
		"""Send filtered motion state to OpenHAB item."""
		target_state = "ON" if self.state in {"Unlocked_Motion", "Unlocked_MotionExtended"} else "OFF"
		if target_state != self._item_motion_filtered.value:
			self._item_motion_filtered.oh_send_command(target_state)

	def _raw_motion_active(self) -> bool:
		"""Check if raw motion is active

		:return: True if active, else False
		"""
		return bool(self._item_motion_raw)

	def _brightness_over_threshold(self) -> bool:
		"""Check if brightness is over threshold

		:return: True if active, else False
		"""
		return self._hysteresis_switch.get_output(self._item_brightness.value)

	def _motion_extended_configured(self) -> bool:
		"""Check if extended motion is configured

		:return: True if active, else False
		"""
		return self._timeout_extended_motion > 0

	def _post_sleep_lock_configured(self) -> bool:
		"""Check if post sleep lock is configured

		:return: True if active, else False
		"""
		return self._timeout_post_sleep_lock > 0

	def _sleep_active(self) -> bool:
		"""Check if sleeping is active

		:return: True if sleeping is active, else False
		"""
		return self._item_sleep.value == habapp_rules.system.SleepState.SLEEPING.value

	def _get_brightness_threshold(self) -> int:
		"""Get the current brightness threshold value.
		
		:return: brightness threshold
		:raises habapp_rules.core.exceptions.HabAppRulesException: if brightness value not given by item or value
		"""
		if self._brightness_threshold_value:
			return self._brightness_threshold_value
		if self._item_brightness_threshold is not None:
			return value if (value := self._item_brightness_threshold.value) else float("inf")
		raise habapp_rules.core.exceptions.HabAppRulesException(f"Can not get brightness threshold. Brightness value or item is not given. value: {self._brightness_threshold_value} | item: {self._item_brightness}")

	# pylint: disable=invalid-name
	def on_enter_Unlocked_Init(self):
		"""Callback, which is called on enter of Unlocked_Init state"""
		if self._item_brightness is not None and self._brightness_over_threshold():
			self.to_Unlocked_TooBright()
		elif self._item_motion_raw.is_on():
			self.to_Unlocked_Motion()
		else:
			self.to_Unlocked_Wait()

	def _check_brightness(self, value: float | None = None) -> None:
		"""Check if brightness is higher than the threshold and trigger the class methods.

		:param value: Value to check. None if last value should be used
		"""
		if self._hysteresis_switch.get_output(value):
			self.brightness_over_threshold()
		else:
			self.brightness_below_threshold()

	def _cb_threshold_change(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the brightness threshold state changed.

		:param event: trigger event
		"""
		self._hysteresis_switch.set_threshold_on(event.value)
		self._check_brightness()

	def _cb_motion_raw(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the raw motion state changed.

		:param event: trigger event
		"""
		if event.value == "ON":
			self.motion_on()
		else:
			self.motion_off()

	def _cb_brightness(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the brightness state changed.

		:param event: trigger event
		"""
		self._check_brightness(event.value)

	def _cb_lock(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the lock state changed.

		:param event: trigger event
		"""
		if event.value == "ON":
			self.lock_on()
		else:
			self.lock_off()

	def _cb_sleep(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Callback, which is triggered if the sleep state changed.

		:param event: trigger event
		"""
		if event.value == habapp_rules.system.SleepState.SLEEPING.value:
			self.sleep_started()
		if event.value == habapp_rules.system.SleepState.AWAKE.value:
			self.sleep_end()



"""Get type of day."""
import datetime

import holidays

_BY_HOLIDAYS = holidays.country_holidays("DE", "BY")


def is_weekend(offset_days: int = 0) -> bool:
	"""Check if current (or offset day) is a weekend day

	:param offset_days: 0 means today, 1 tomorrow, -1 yesterday
	:return: True if desired day is a weekend day
	"""
	day_to_check = datetime.datetime.today() + datetime.timedelta(days=offset_days)
	return day_to_check.isoweekday() in {6, 7}


def is_holiday(offset_days: int = 0) -> bool:
	"""Check if current (or offset day) is holiday

	:param offset_days: 0 means today, 1 tomorrow, -1 yesterday
	:return: True if desired day is holiday
	"""
	day_to_check = datetime.datetime.today() + datetime.timedelta(days=offset_days)
	return day_to_check in _BY_HOLIDAYS



"""Rule to detect whether it is summer or winter."""
import datetime
import logging
import statistics

import HABApp

import habapp_rules.common.hysteresis
import habapp_rules.core.logger

LOGGER = logging.getLogger(__name__)


class SummerWinterException(Exception):
	"""Custom Exception for SummerWinter"""


class SummerWinter(HABApp.Rule):
	"""Rule check if it is summer or winter."""

	def __init__(self, outside_temperature_name: str, summer_name: str, persistence_service: str = None, days: int = 5, temperature_threshold: float = 16, last_check_name: str = None) -> None:
		"""Init rule to update summer/winter item.

		:param outside_temperature_name: Name of outside temperature item. OpenHAB-Type must be Number item
		:param summer_name: Name of summer item. OpenHAB-Type must be Switch item
		:param persistence_service: Name of persistence service
		:param days: number of days in the past which will be used to check if it is summer
		:param temperature_threshold: threshold weighted temperature for summer
		:param last_check_name: Name of last check item. OpenHAB-Type must be DateTime item
		"""
		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, summer_name)

		# set class variables
		self._persistence_service = persistence_service
		self._days = days
		self._hysteresis_switch = habapp_rules.common.hysteresis.HysteresisSwitch(temperature_threshold, 0.5)
		self.__now = datetime.datetime.now()

		# get items
		self._outside_temp_item = HABApp.openhab.items.NumberItem.get_item(outside_temperature_name)
		self._item_summer = HABApp.openhab.items.SwitchItem.get_item(summer_name)
		self._item_last_check = HABApp.openhab.items.DatetimeItem.get_item(last_check_name) if last_check_name else None

		# run at init and every day at 23:00
		self.run.soon(self._cb_update_summer)
		self.run.on_every_day(datetime.time(23), self._cb_update_summer)

		LOGGER.debug("Init of Summer / Winter successful")

	def __get_weighted_mean(self, days_in_past: int) -> float:
		"""Get weighted mean temperature.

		The weighted mean temperature will be calculated according the following formula: (T07 + T14 + T22 + T22) / 4 where T07 is the temperature at 7:00 (and so on)
		It is possible to get the weighted temperature of today or of some days in the past -> defined by the days_in past attribute. If this method is called before 22:00 there will be an offset of one day.
		:param days_in_past: if days in past is set to 0 it will return the mean of today. 1 will return the offset of yesterday
		:return: the weighted mean according the formula in doc-string
		:raises SummerWinterException: if there is not enough data for at least one evaluated hour.
		"""
		day_offset = 0
		if self.__now.hour < 23:
			day_offset = 1

		temperature_values = []
		for hour in [7, 14, 22]:
			start_time = datetime.datetime(self.__now.year, self.__now.month, self.__now.day, hour, 0) - datetime.timedelta(days=day_offset + days_in_past)
			end_time = start_time + datetime.timedelta(hours=1)
			persistence_data = self._outside_temp_item.get_persistence_data(persistence=self._persistence_service, start_time=start_time, end_time=end_time)
			if not persistence_data.data:
				raise SummerWinterException(f"No data for {start_time}")
			temperature_values.append(next(iter(persistence_data.data.values())))
		return (sum(temperature_values) + temperature_values[2]) / 4

	def __is_summer(self) -> bool:
		"""Check if it is summer (or winter).

		:return: Returns True if it is summer.
		:raises SummerWinterException: if summer/winter could not be detected
		"""
		self.__now = datetime.datetime.now()
		values = []
		for day in range(self._days):
			try:
				values.append(self.__get_weighted_mean(day))
			except SummerWinterException:
				self._instance_logger.warning(f"Could not get mean value of day -{day}")

		if not values:
			raise SummerWinterException("Not enough data to detect summer/winter")

		is_summer = self._hysteresis_switch.get_output(mean_value := statistics.mean(values))
		self._instance_logger.debug(f"Check Summer/Winter. values = {values} | mean = {mean_value} | summer = {is_summer}")
		return is_summer

	def _cb_update_summer(self) -> None:
		"""Callback to update the summer item."""
		try:
			is_summer = self.__is_summer()
		except SummerWinterException:
			self._instance_logger.error("Could not get summer / winter")
			return

		# get target state of summer
		target_value = "ON" if is_summer else "OFF"

		# send state
		if self._item_summer.value != target_value:
			self._item_summer.oh_send_command(target_value)
			self._instance_logger.info(f"Summer changed to {target_value}")

		# update last update item at every call
		if self._item_last_check:
			self._item_last_check.oh_send_command(datetime.datetime.now())



"""Implementations of logical functions."""
import abc
import logging

import HABApp

import habapp_rules.core.helper
import habapp_rules.core.logger

LOGGER = logging.getLogger(__name__)


class _BinaryLogicBase(HABApp.Rule):
	"""Base class for binary logical functions."""

	def __init__(self, input_names: list[str], output_name: str) -> None:
		"""Init a logical function.

		:param input_names: list of input items (must be either Switch or Contact and all have to match to output_item)
		:param output_name: name of output item
		:raises TypeError: if unsupported item-type is given for output_name
		"""
		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, f"{self.__class__.__name__}_{output_name}")

		self._output_item = HABApp.openhab.items.OpenhabItem.get_item(output_name)

		if isinstance(self._output_item, HABApp.openhab.items.SwitchItem):
			self._positive_state = "ON"
			self._negative_state = "OFF"
		elif isinstance(self._output_item, HABApp.openhab.items.ContactItem):
			self._positive_state = "CLOSED"
			self._negative_state = "OPEN"
		else:
			raise TypeError(f"Item type '{type(self._output_item)}' is not supported. Type must be SwitchItem or ContactItem")

		self._input_items = []
		for name in input_names:
			if isinstance(input_item := HABApp.openhab.items.OpenhabItem.get_item(name), type(self._output_item)):
				self._input_items.append(input_item)
				input_item.listen_event(self._cb_input_event, HABApp.openhab.events.ItemStateUpdatedEventFilter())
			else:
				self._instance_logger.error(f"Item '{name}' must have the same type like the output item. Expected: {type(self._output_item)} | actual : {type(input_item)}")

		self._cb_input_event(None)
		self._instance_logger.debug(f"Init of rule '{self.__class__.__name__}' with was successful. Output item = '{output_name}' | Input items = '{input_names}'")

	@abc.abstractmethod
	def _cb_input_event(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | None) -> None:
		"""Callback, which is called if one of the input items had a state event.

		:param event: item event of the updated item
		"""

	def _set_output_state(self, output_state: str) -> None:
		"""Set state to the output element

		:param output_state: state which will be set
		"""
		if isinstance(self._output_item, HABApp.openhab.items.ContactItem):
			self._output_item.oh_post_update(output_state)
		else:
			habapp_rules.core.helper.send_if_different(self._output_item, output_state)


class And(_BinaryLogicBase):
	"""Logical AND function.

	Example:
	habapp_rules.common.logic.And(["Item_1", "Item_2"], "Item_result")
	"""

	def _cb_input_event(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | None) -> None:
		"""Callback, which is called if one of the input items had a state event.

		:param event: item event of the updated item
		"""
		output_state = self._positive_state if all(item.value == self._positive_state for item in self._input_items) else self._negative_state
		self._set_output_state(output_state)


class Or(_BinaryLogicBase):
	"""Logical OR function.

	Example:
	habapp_rules.common.logic.Or(["Item_1", "Item_2"], "Item_result")
	"""

	def _cb_input_event(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | None) -> None:
		"""Callback, which is called if one of the input items had a state event.

		:param event: item event of the updated item
		"""
		output_state = self._positive_state if any(item.value == self._positive_state for item in self._input_items) else self._negative_state
		self._set_output_state(output_state)


class _NumericLogicBase(HABApp.Rule):
	"""Base class for numeric logical functions."""

	def __init__(self, input_names: list[str], output_name: str, ignore_old_values_time: int | None = None) -> None:
		"""Init a logical function.

		:param input_names: list of input items (must be either Dimmer or Number and all have to match to output_item)
		:param output_name: name of output item
		:param ignore_old_values_time: ignores values which are older than the given time in seconds. If None, all values will be taken
		:raises TypeError: if unsupported item-type is given for output_name
		"""
		self._ignore_old_values_time = ignore_old_values_time

		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, f"{self.__class__.__name__}_{output_name}")

		self._output_item = HABApp.openhab.items.OpenhabItem.get_item(output_name)

		if not isinstance(self._output_item, (HABApp.openhab.items.DimmerItem, HABApp.openhab.items.NumberItem)):
			raise TypeError(f"Item type '{type(self._output_item)}' is not supported. Type must be NumberItem or DimmerItem")

		self._input_items = []
		for name in input_names:
			if isinstance(input_item := HABApp.openhab.items.OpenhabItem.get_item(name), type(self._output_item)):
				self._input_items.append(input_item)
				input_item.listen_event(self._cb_input_event, HABApp.openhab.events.ItemStateChangedEventFilter())
			else:
				self._instance_logger.error(f"Item '{name}' must have the same type like the output item. Expected: {type(self._output_item)} | actual : {type(input_item)}")

		self._cb_input_event(None)
		self._instance_logger.debug(f"Init of rule '{self.__class__.__name__}' with was successful. Output item = '{output_name}' | Input items = '{input_names}'")

	def _cb_input_event(self, event: HABApp.openhab.events.ItemStateUpdatedEvent | None) -> None:
		"""Callback, which is called if one of the input items had a state event.

		:param event: item event of the updated item
		"""
		filtered_items = habapp_rules.core.helper.filter_updated_items(self._input_items, self._ignore_old_values_time)
		value = self._apply_numeric_logic([item.value for item in filtered_items if item is not None])

		if value is None:
			return

		self._set_output_state(value)

	@abc.abstractmethod
	def _apply_numeric_logic(self, input_values: list[float]) -> float:
		"""Apply numeric logic

		:param input_values: input values
		:return: value which fulfills the filter type
		"""

	def _set_output_state(self, output_state: str) -> None:
		"""Set state to the output element

		:param output_state: state which will be set
		"""
		habapp_rules.core.helper.send_if_different(self._output_item, output_state)


class Min(_NumericLogicBase):
	"""Logical Min function with filter for old / not updated items.

	Example:
	habapp_rules.common.logic.Min(["Item_1", "Item_2"], "Item_result", 600)
	"""

	def _apply_numeric_logic(self, input_values: list[float]) -> float:
		"""Apply numeric logic

		:param input_values: input values
		:return: min value of the given values
		"""
		return HABApp.util.functions.min(input_values)


class Max(_NumericLogicBase):
	"""Logical Max function with filter for old / not updated items.

	Example:
	habapp_rules.common.logic.Max(["Item_1", "Item_2"], "Item_result", 600)
	"""

	def _apply_numeric_logic(self, input_values: list[float]) -> float:
		"""Apply numeric logic

		:param input_values: input values
		:return: max value of the given values
		"""
		return HABApp.util.functions.max(input_values)


class Sum(_NumericLogicBase):
	"""Logical Sum function with filter for old / not updated items.

	Example:
	habapp_rules.common.logic.Sum(["Item_1", "Item_2"], "Item_result", 600)
	"""

	def __init__(self, input_names: list[str], output_name: str, ignore_old_values_time: int | None = None) -> None:
		"""Init a logical function.

		:param input_names: list of input items (must be either Dimmer or Number and all have to match to output_item)
		:param output_name: name of output item
		:param ignore_old_values_time: ignores values which are older than the given time in seconds. If None, all values will be taken
		:raises TypeError: if unsupported item-type is given for output_name
		"""
		_NumericLogicBase.__init__(self, input_names, output_name, ignore_old_values_time)
		if isinstance(self._output_item, HABApp.openhab.items.DimmerItem):
			raise TypeError(f"Dimmer items can not be used for Sum function! Given input_names: {input_names} | output_name: {output_name}")

	def _apply_numeric_logic(self, input_values: list[float]) -> float:
		"""Apply numeric logic

		:param input_values: input values
		:return: min value of the given values
		"""
		return sum(val for val in input_values if val is not None)


class InvertValue(HABApp.Rule):
	"""Rule to update another item if the value of an item changed.

	Example:
	habapp_rules.common.logic.InvertValue("Item_1", "Item_2")
	"""

	def __init__(self, input_name: str, output_name: str, only_positive: bool = False, only_negative: bool = False) -> None:
		"""Init rule.

		:param input_name: Name of input item (NumberItem)
		:param output_name: Name of output item (NumberItem)
		:param only_positive: if true, only positive values will be set to output item
		:param only_negative: if true, only negative values will be set to output item
		"""
		HABApp.Rule.__init__(self)
		self._instance_logger = habapp_rules.core.logger.InstanceLogger(LOGGER, f"{self.__class__.__name__}_{output_name}")

		self._item_input = HABApp.openhab.items.NumberItem.get_item(input_name)
		self._item_output = HABApp.openhab.items.NumberItem.get_item(output_name)
		self._only_positive = only_positive
		self._only_negative = only_negative

		self._item_input.listen_event(self._cb_input_value, HABApp.openhab.events.ItemStateChangedEventFilter())
		self._cb_input_value(HABApp.openhab.events.ItemStateChangedEvent(self._item_input.name, self._item_input.value, None))
		self._instance_logger.debug(f"Init of rule '{self.__class__.__name__}' with was successful. Output item = '{output_name}' | Input item = '{input_name}'")

	def _cb_input_value(self, event: HABApp.openhab.events.ItemStateChangedEvent) -> None:
		"""Set output, when input value changed

		:param event: event, which triggered this callback
		"""
		if event.value is None:
			return

		output_value = -1 * event.value

		if self._only_negative and output_value > 0:
			output_value = 0
		elif self._only_positive and output_value < 0:
			output_value = 0

		self._item_output.oh_send_command(output_value)
